﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Threading;

namespace Ascend
{
	/// <summary>
	/// Class for specifying an AnimatedSceneNode's animation within a Timeline.
	/// </summary>
	public class TimelineAnimation
	{
		#region Lifecycle
		/// <summary>
		/// Creates a new TimelineAnimation instance with given AnimatedSceneNode,
		/// animation name, and start frame.
		/// </summary>
		/// <param name="sceneNode">The AnimatedSceneNode to be animated.</param>
		/// <param name="animationName">The animation to play.</param>
		/// <param name="start">The Timeline frame at which to begin the animation.</param>
		/// <param name="relativePlayRate">The non-zero ratio of frames to progress for every one Timeline frame.
		/// Specifying a negative value will reverse the animation.</param>
		/// <exception cref="ArgumentException">
        /// Thrown if <paramref name="animationName"/> is not a valid animation defined on <paramref name="sceneNode"/>,
        /// if <paramref name="start"/> is less than zero,
        /// or if <paramref name="relativePlayRate"/> is zero.
		/// </exception>
		public TimelineAnimation(AnimatedSceneNode sceneNode, string animationName, int start = 0, double relativePlayRate = 1.0)
		{
			Initialize(sceneNode, animationName, relativePlayRate);
			SetFrameRange(start);
		}

		/// <summary>
		/// Creates a new TimelineAnimation instance with given AnimatedSceneNode,
		/// animation name, start frame, and end frame. If the length of the animation
		/// defined on the AnimatedSceneNode is less than the specified frame range,
		/// the animation will loop until <paramref name="end"/> is reached.
		/// </summary>
		/// <param name="sceneNode">The AnimatedSceneNode to be animated.</param>
		/// <param name="animationName">The animation to play.</param>
		/// <param name="start">The Timeline frame at which to begin the animation.</param>
		/// <param name="end">The Timeline frame at which to end the animation.</param>
		/// <param name="relativePlayRate">The non-zero ratio of frames to progress for every one Timeline frame.
		/// Specifying a negative value will reverse the animation.</param>
		/// <exception cref="ArgumentException">
        /// Thrown if <paramref name="animationName"/> is not a valid animation defined on <paramref name="sceneNode"/>,
        /// if <paramref name="start"/> is less than zero or greater than <paramref name="end"/>,
        /// or if <paramref name="relativePlayRate"/> is zero.
        /// </exception>
		public TimelineAnimation(AnimatedSceneNode sceneNode, string animationName, int start, int end, double relativePlayRate = 1.0)
		{
			Initialize(sceneNode, animationName, relativePlayRate);
			SetFrameRange(start, end);
		}
		#endregion

		#region Internal Members
		internal List<AnimationFrame> _frames;
		#endregion

		#region Properties
		/// <summary>
		/// Gets the AnimatedSceneNode to be animated.
		/// </summary>
		public AnimatedSceneNode SceneNode { get; private set; }

		/// <summary>
		/// Gets the name of the animation to play.
		/// </summary>
		public string AnimationName { get; private set; }

		/// <summary>
		/// Gets the start Timeline frame for the AnimatedSceneNode's animation,
		/// since the start of the Timeline.
		/// </summary>
		public int Start { get; private set; }

		/// <summary>
		/// Gets the end Timeline frame for the AnimatedSceneNode's animation,
		/// since the start of the Timeline.
		/// </summary>
		public int End { get; private set; }

		/// <summary>
		/// Gets the value specifying the ratio of frames to progress for
		/// every one Timeline frame.
		/// </summary>
		public double RelativePlayRate { get; private set; }
		#endregion

		#region Private Methods
		private void Initialize(AnimatedSceneNode sceneNode, string animationName, double relativePlayRate)
		{
			if (!sceneNode.ListAnimations().Contains(animationName))
				throw new ArgumentException(string.Format("Animation '{0}' does not exist on AnimatedSceneNode.", animationName));

			if (relativePlayRate == 0.0)
				throw new ArgumentException("Relative play rate must be a non-zero value.");

			SceneNode = sceneNode;
			AnimationName = animationName;
			RelativePlayRate = relativePlayRate;
		}

		private void SetFrameRange(int start, int end = -1)
		{
			if (start < 0)
				throw new ArgumentException("Start frame of animation must be non-negative.");

			if (end != -1 && start > end)
				throw new ArgumentException("Start frame cannot be greater than end frame.");

			Start = start;

			FrameInterval interval = SceneNode.GetAnimationInterval(AnimationName);

			if (end == -1)
			{
				// Use native animation length (plays through animation once)
				_frames = SceneNode.AnimationFrames.GetRange(interval.Start, interval.Length);
				End = start + ToTimelineLength(interval.Length) - 1;
			}
			else
			{
				// Use specified run length, up to total number of frames in animation
				int runLength = ToLocalAnimationLength(end - start + 1);
				
				_frames = SceneNode.AnimationFrames.GetRange(interval.Start, Math.Min(runLength, interval.Length));
				End = end;
			}

			if (RelativePlayRate < 0.0)
				_frames.Reverse();
		}

		private int ToTimelineLength(int localAnimationLength)
		{
			return (int)Math.Ceiling(localAnimationLength / Math.Abs(RelativePlayRate));
		}

		private int ToLocalAnimationLength(int timelineLength)
		{
			return (int)Math.Floor(timelineLength * Math.Abs(RelativePlayRate));
		}
		#endregion
	}


	/// <summary>
	/// Class for creating and playing multiple synchronized animations.
	/// </summary>
	public class Timeline : DispatcherObject
	{
		#region Lifecycle
		/// <summary>
		/// Creates a new Timeline instance.
		/// </summary>
		public Timeline()
		{
			Animations = new ObservableCollection<TimelineAnimation>();
			Animations.CollectionChanged += Animations_CollectionChanged;
			AnimationState = AnimationState.Stopped;
		}
		#endregion

		#region Events
		/// <summary>
		/// Occurs when a non-looping animation completes.
		/// </summary>
		public event AnimationEventHandler Completed;
		#endregion

		#region Private Members
		private object _animLock = new object();
		#endregion

		#region Properties
		private int _currentFrameNumber;
		/// <summary>
		/// Gets or sets the current frame number for this Timeline.
		/// When setting, value must be between 0 and EndFrameNumber, inclusive.
		/// </summary>
		public int CurrentFrameNumber
		{
			get { return _currentFrameNumber; }
			set
			{
				if (value < 0 || value > EndFrameNumber)
					throw new ArgumentOutOfRangeException("Timeline CurrentFrameNumber property can only be set to a value between 0 and " + EndFrameNumber + ".");

				_currentFrameNumber = value;

				ApplyCurrentFrame();
			}
		}

		/// <summary>
		/// Gets the maximum frame number for this Timeline.
		/// </summary>
		public int EndFrameNumber { get; private set; }

		/// <summary>
		/// Gets the list of animations defined in this Timeline.
		/// </summary>
		public ObservableCollection<TimelineAnimation> Animations { get; private set; }

		/// <summary>
		/// Gets this Timeline's current animation state.
		/// </summary>
		public AnimationState AnimationState { get; private set; }
		#endregion

		#region Public Methods
		/// <summary>
		/// Plays this Timeline.
		/// </summary>
		/// <param name="loop">Whether or not to loop the animation.</param>
		/// <param name="restart">Whether or not to start the Timeline from the beginning.</param>
		/// <param name="fps">Animation speed in frames per second. Specifying a negative value
		/// will play animation in reverse.</param>
        /// <exception cref="ArgumentException">
        /// Thrown if <paramref name="fps"/> is zero.
        /// </exception>
		public void Play(bool loop = false, bool restart = false, double fps = 60.0)
		{
			if (AnimationState == AnimationState.Playing)
				Pause();

            if (fps == 0.0)
                throw new ArgumentException("Frame rate must be a non-zero value.");

			bool reverse = fps < 0.0;

			if (restart || CurrentFrameNumber < 0 || CurrentFrameNumber > EndFrameNumber)
				_currentFrameNumber = reverse ? EndFrameNumber + 1 : -1;  // When animation starts, next frame will be the correct animation bound

			StartAnimationThread(loop, reverse, fps);
		}

		/// <summary>
		/// Pauses this Timeline.
		/// </summary>
		public void Pause()
		{
			AnimationState = AnimationState.Stopped;

			foreach (TimelineAnimation anim in Animations)
				anim.SceneNode.Pause();
		}
		#endregion

		#region Private Methods
		private void StartAnimationThread(bool loop, bool reverse, double fps)
		{
			Task.Factory.StartNew(() =>
				{
					lock (_animLock)
					{
						AnimationState = AnimationState.Playing;

						// Start timing
						DateTime start = DateTime.Now;
						TimeSpan dt;
						int lastFrame = 0;

						while (AnimationState == AnimationState.Playing)
						{
							// Get time and frame deltas
							DateTime thisTime = DateTime.Now;
							dt = thisTime - start;

							int thisFrame = (int)(Math.Abs(fps) * dt.TotalSeconds);
							int df = thisFrame - lastFrame;

							lastFrame = thisFrame;

							if (df >= 1)
							{
								if (ProgressGlobalAnimation(df, reverse, loop))
								{
									RaiseCompletedEvent();
									break;
								}
							}

							// Sleep to rest CPU
							Thread.Sleep((int)Math.Max(5, 1000 / fps - (DateTime.Now - thisTime).Milliseconds - 5));
						}
					}
				}
			);
		}

		private bool ProgressGlobalAnimation(int frameDelta, bool reverse, bool loop)
		{
			int next = reverse ? CurrentFrameNumber - frameDelta : CurrentFrameNumber + frameDelta;

			if ((next <= 0 && reverse) || (next >= EndFrameNumber && !reverse))
			{
				if (!loop)
				{
					// Clamp to end of animation
					Dispatcher.Invoke(() => CurrentFrameNumber = reverse ? 0 : EndFrameNumber);
					
					Pause();

					// Flag that the timeline completed
					return true;
				}
				
				// Compute wrapped-around frame number
				next = (next < 0)
					? (next % (EndFrameNumber + 1)) + EndFrameNumber + 1
					: next % (EndFrameNumber + 1);
			}

			Dispatcher.Invoke(() => CurrentFrameNumber = next);

			return false;
		}

		private void ApplyCurrentFrame()
		{
			foreach (var anim in Animations)
			{
				if (CurrentFrameNumber >= anim.Start && CurrentFrameNumber <= anim.End)
				{
					int idx = (int)((CurrentFrameNumber - anim.Start) * Math.Abs(anim.RelativePlayRate)) %
						anim._frames.Count;
					
					anim.SceneNode.CurrentFrame = anim._frames[idx];

					if (AnimationState == AnimationState.Playing)
					{
						anim.SceneNode.AnimationState = AnimationState.PlayingInTimeline;
						anim.SceneNode.CurrentAnimation = anim.AnimationName;
					}
				}
				else if (AnimationState == AnimationState.Playing && 
					!Animations.Any(a => a.SceneNode == anim.SceneNode &&
					CurrentFrameNumber >= a.Start && CurrentFrameNumber <= a.End))
				{
					// If node is not being animated at this frame by another item in Animations,
					// set this node's state to stopped
					anim.SceneNode.AnimationState = AnimationState.Stopped;
					anim.SceneNode.CurrentAnimation = null;
				}
			}
		}
		#endregion

		#region Event Handlers
		private void Animations_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
		{
			if (AnimationState == AnimationState.Playing)
				throw new InvalidOperationException("Cannot edit TimelineAnimation collection while Timeline is playing.");

			// Check for overlapping animations
			foreach (var newAnim in e.NewItems.Cast<TimelineAnimation>())
			{
				foreach (var currAnim in Animations.Where(a => a != newAnim && a.SceneNode == newAnim.SceneNode))
				{
					if ((newAnim.Start >= currAnim.Start && newAnim.Start <= currAnim.End) ||
						(currAnim.Start >= newAnim.Start && currAnim.Start <= newAnim.End))
						throw new InvalidOperationException("Added TimelineAnimation overlaps an existing TimelineAnimation for the same AnimatedSceneNode.");
				}
			}

			// Recompute end frame
			EndFrameNumber = (Animations.Count > 0) ? Animations.Max(a => a.End) : 0;

			// In case an animation was removed, check current frame number
			if (CurrentFrameNumber > EndFrameNumber)
				CurrentFrameNumber = EndFrameNumber;
		}
		#endregion

		#region Event Invokers
		#region Completed
		private void RaiseCompletedEvent()
		{
			if (Completed != null)
			{
				AnimationEventArgs e = new AnimationEventArgs();

				foreach (var del in Completed.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion
		#endregion
	}
}