﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media.Media3D;

namespace Ascend
{
	/// <summary>
	/// Class that describes the skeletal structure for a ModelSceneNode.
	/// </summary>
	public class Armature : DependencyObject, ICloneable
	{
		#region Inner Classes
		private struct WeightedTransform
		{
			public WeightedTransform(double weight, Transform3D transform)
			{
				Weight = weight;
				Transform = transform;
			}

			public double Weight;
			public Transform3D Transform;
		}
		#endregion

		#region Lifecycle
		/// <summary>
		/// Creates a default Armature instance.
		/// </summary>
		public Armature()
		{
			Data = new Model3DGroup();

			// Create root Bone model group
			Data.Children.Add(new Model3DGroup());

			// Create root Bone collection. Each element will correspond
			// to a model in the root Bone model group.
			RootBones = new BoneCollection();

			// Listen to changes in root Bones
			RootBones.CollectionChanged += CollectionChanged;
		}
		#endregion

		#region Private Members
		private Dictionary<int, List<WeightedTransform>> _boneTransforms = new Dictionary<int, List<WeightedTransform>>();

		internal Model3DGroup Data
		{
			get { return (Model3DGroup)GetValue(DataProperty); }
			set { SetValue(DataProperty, value); }
		}

		private static readonly DependencyProperty DataProperty =
			DependencyProperty.Register("Data", typeof(Model3DGroup), typeof(Armature));


		private Model3DCollection VisualChildren
		{
			get { return ((Model3DGroup)Data.Children[0]).Children; }
		}
		#endregion

		#region Properties
		/// <summary>
		/// Gets or sets the name of this Armature.
		/// </summary>
		public string Name { get; set; }

		private AnimatedSceneNode _skin;
		/// <summary>
		/// Gets the AnimatedSceneNode acting as this Armature's skin.
		/// </summary>
		public AnimatedSceneNode Skin
		{
			get { return _skin; }

			internal set
			{
				if (_skin == value)
					return;

				_skin = value;

				if (_skin != null)
					_skin.Armature = this;
			}
		}

		/// <summary>
		/// Gets this Armature's collection of root Bones.
		/// </summary>
		public BoneCollection RootBones { get; private set; }
		#endregion

		#region Public Methods
		/// <summary>
		/// Places each Bone into its rest pose and updates the Armature's skin.
		/// </summary>
		public void Rest()
		{
			foreach (var bone in ListBones())
				bone.Rest();

			UpdateSkin();
		}

		/// <summary>
		/// Updates this Armature's skin by applying its current pose.
		/// </summary>
		public void UpdateSkin()
		{
			if (RootBones == null || Skin == null || Skin.Model == null || Skin.Model.Geometry is MeshGeometry3D == false)
				return;

			int idx = RemoveSkinFromSceneGraph();

			MeshGeometry3D mesh = (MeshGeometry3D)Skin.Model.Geometry;

			foreach (var kvp in _boneTransforms)
			{
				int vert = kvp.Key;
				bool hasNormal = (Skin.RestNormals.Count > vert);
				Point3D inPoint = Skin.RestVertexPositions[vert];
				Vector3D inNormal = hasNormal ? Skin.RestNormals[vert] : default(Vector3D);
				Point3D outPoint = new Point3D();
				Vector3D outNormal = new Vector3D();
				double weightSum = 0.0;

				foreach (var trans in kvp.Value)
				{
					outPoint += trans.Weight * (Vector3D)Bone.TransformSkinPoint(trans.Transform, inPoint);

					if (hasNormal)
						outNormal += trans.Weight * Bone.TransformSkinNormal(trans.Transform, inNormal);

					weightSum += trans.Weight;
				}

				outPoint = (Point3D)(((Vector3D)outPoint) / weightSum);
				mesh.Positions[vert] = outPoint;

				if (hasNormal)
				{
					outNormal /= weightSum;
					outNormal.Normalize();
					mesh.Normals[vert] = outNormal;
				}
			}

			if (idx != -1)
				RestoreSkinToSceneGraph(idx);
		}

		/// <summary>
		/// Creates a deep copy of this Armature. Since an AnimatedSceneNode
		/// may only have one Armature, the clone's Skin property will be null.
		/// </summary>
		public object Clone()
		{
			Armature clone = new Armature { Name = Name };

			for (int c = 0; c < RootBones.Count(); c++)
				clone.RootBones.Add(RootBones[c].Clone() as Bone);

			return clone;
		}

		/// <summary>
		/// Returns this Armature as a string.
		/// </summary>
		public override string ToString()
		{
			return (string.IsNullOrWhiteSpace(Name) ? string.Empty : Name + " ")
				+ "(" + GetType().Name + ")";
		}
		#endregion

		#region Internal Methods
		internal List<Bone> ListBones()
		{
			List<Bone> bones = new List<Bone>();

			foreach (var bone in RootBones)
				ListBonesRecursive(bone, bones);

			return bones;
		}

		internal void UpdateBoneTransforms()
		{
			_boneTransforms = new Dictionary<int, List<WeightedTransform>>();

			foreach (var bone in ListBones())
			{
				foreach (var inf in bone.VertexInfluences.Where(i => i.Weight != 0.0))
				{
					foreach (int v in inf.Vertices)
					{
						WeightedTransform trans = new WeightedTransform(inf.Weight, bone.ComputeSkinTransform());

						if (_boneTransforms.ContainsKey(v))
							_boneTransforms[v].Add(trans);
						else
							_boneTransforms.Add(v, new List<WeightedTransform> { trans });
					}
				}
			}
		}

		internal void ApplyPoses(List<ObjectPose> bonePoses)
		{
			List<Bone> bones = ListBones();

			foreach (var pose in bonePoses)
			{
				Bone bone = bones[bonePoses.IndexOf(pose)];

				if (pose != null)
				{
					bone.X = pose.X;
					bone.Y = pose.Y;
					bone.Z = pose.Z;
					bone.AngleX = pose.AngleX;
					bone.AngleY = pose.AngleY;
					bone.AngleZ = pose.AngleZ;
					bone.ScaleX = pose.ScaleX;
					bone.ScaleY = pose.ScaleY;
					bone.ScaleZ = pose.ScaleZ;
				}
			}

			UpdateSkin();
		}
		#endregion

		#region Private Methods
		private int RemoveSkinFromSceneGraph()
		{
			int idx = -1;

			if (Skin.TransformParent is SceneNode)
			{
				idx = ((SceneNode)Skin.Parent).VisualChildren.IndexOf(Skin.Data);
				((SceneNode)Skin.Parent).VisualChildren.RemoveAt(idx);
			}
			else if (Skin.TransformParent is Bone)
			{
				idx = Skin.BoneParent.AttachedVisualChildren.IndexOf(Skin.Data);
				Skin.BoneParent.AttachedVisualChildren.RemoveAt(idx);
			}

			return idx;
		}

		private void RestoreSkinToSceneGraph(int index)
		{
			if (Skin.TransformParent is SceneNode)
				((SceneNode)Skin.Parent).VisualChildren.Insert(index, Skin.Data);
			else
				Skin.BoneParent.AttachedVisualChildren.Insert(index, Skin.Data);
		}

		private void ListBonesRecursive(Bone bone, List<Bone> bones)
		{
			bones.Add(bone);

			foreach (var child in bone.Children)
				ListBonesRecursive(child, bones);
		}
		#endregion

		#region Event Handlers
		private void CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
		{
			switch (e.Action)
			{
				case NotifyCollectionChangedAction.Add:
					HandleCollectionActionAdd(sender, e);
					break;

				case NotifyCollectionChangedAction.Remove:
					HandleCollectionActionRemove(sender, e);
					break;

				default:
					throw new NotSupportedException("NotifyCollectionChangedAction.Reset is not supported.");
			}

			UpdateBoneTransforms();
		}

		private void HandleCollectionActionAdd(object sender, NotifyCollectionChangedEventArgs e)
		{
			foreach (Bone bone in e.NewItems)
			{
				// Add Bone's data to root Bone model group
				VisualChildren.Add(bone.Data);

				// Remove Bone from any current transform parent
				if (bone.Parent != null)
					bone.Parent.Children.Remove(bone);  // Will set bone.Parent = null
				else if (bone.Armature != null)
					bone._armature = null;

				// Set armature
				bone._armature = this;
			}
		}

		private void HandleCollectionActionRemove(object sender, NotifyCollectionChangedEventArgs e)
		{
			foreach (Bone bone in e.OldItems)
			{
				// Remove Bone's data from model group
				VisualChildren.Remove(bone.Data);

				// Clear armature
				bone._armature = null;
			}
		}
		#endregion
	}
}
