﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media.Media3D;

namespace Ascend
{
	/// <summary>
	/// Class for defining vertex influences within an Armature.
	/// </summary>
	public class Bone : TransformableObject
	{
		#region Lifecycle
		/// <summary>
		/// Creates a default Bone instance.
		/// </summary>
		public Bone()
		{
			// Create children model group
			Data.Children.Add(new Model3DGroup());

			// Create child Bone collection. Each element will correspond
			// to a model in the children model group.
			Children = new BoneCollection();

			// Create attached SceneNodes model group
			Data.Children.Add(new Model3DGroup());

			// Create attached SceneNode collection. Each element will
			// correspond to a model in the attached SceneNode model group.
			RigidAttachments = new SceneNodeCollection();

			// Listen to changes in children and attached nodes
			Children.CollectionChanged += Children_CollectionChanged;
			RigidAttachments.CollectionChanged += Children_CollectionChanged;

			VertexInfluences = new ObservableCollection<VertexInfluence>();
			VertexInfluences.CollectionChanged += VertexInfluences_CollectionChanged;

			SetupRestTransforms();
		}
		#endregion

		#region Private Members
		private TranslateTransform3D _translateTransformRest;
		private RotateTransform3D _rotateTransformXRest;
		private RotateTransform3D _rotateTransformYRest;
		private RotateTransform3D _rotateTransformZRest;
		private ScaleTransform3D _scaleTransformRest;
		private Transform3DGroup _restingTransform;

		private Model3DCollection VisualChildren
		{
			get { return ((Model3DGroup)Data.Children[0]).Children; }
		}

		internal Model3DCollection AttachedVisualChildren
		{
			get { return ((Model3DGroup)Data.Children[1]).Children; }
		}
		#endregion

		#region Properties
		internal Armature _armature;

		public Armature Armature
		{
			get
			{
				if (Parent == null)
					return _armature;

				return Parent.Armature;
			}
		}

		private Bone _parent;
		/// <summary>
		/// Gets or sets this Bone's parent.
		/// </summary>
		public Bone Parent
		{
			get { return _parent; }
			set
			{
				// Do nothing in this case
				if (_parent == value)
					return;

				// Remove from curent parent
				if (value == null)
				{
					_parent.Children.Remove(this);  // Will set _parent = null
					return;
				}

				// Attach to new parent (will handle removing from current parent)
				value.Children.Add(this);  // Will set _parent = value
			}
		}

		/// <summary>
		/// Gets the TransformableObject to which this Bone is relatively transformed.
		/// It will be equal to Parent, Armature.Skin, or null.
		/// </summary>
		public override TransformableObject TransformParent
		{
			get
			{
				if (Parent != null)
					return Parent;
				else if (Armature != null && Armature.Skin != null)
					return Armature.Skin;
				else
					return null;
			}
		}

		/// <summary>
		/// Gets a read-only list of TransformableObjects that are transformed relative to this Bone.
		/// </summary>
		public override IList<TransformableObject> TransformChildren
		{
			get
			{
				List<TransformableObject> transChildren = RigidAttachments.Cast<TransformableObject>().Concat(Children.Cast<TransformableObject>()).ToList();
				
				return transChildren.AsReadOnly();
			}
		}

		/// <summary>
		/// Gets this Bone's collection of child Bones.
		/// </summary>
		public BoneCollection Children { get; private set; }

		/// <summary>
		/// Gets this Bone's collection of rigidly-attached SceneNodes.
		/// (Adding a SceneNode to this collection will nullify its Parent property.)
		/// </summary>
		public SceneNodeCollection RigidAttachments { get; private set; }

		/// <summary>
		/// Gets the set of VertexInfluences that describes how this Bone affects its skin.
		/// </summary>
		public ObservableCollection<VertexInfluence> VertexInfluences { get; private set; }

		private double _restX;
		/// <summary>
		/// Gets or sets this Bone's resting X coordinate.
		/// </summary>
		public double RestX
		{
			get { return _restX; }
			set
			{
				if (_restX == value)
					return;

				_restX = value;

				_translateTransformRest.OffsetX = _restX;
			}
		}

		private double _restY;
		/// <summary>
		/// Gets or sets this Bone's resting Y coordinate.
		/// </summary>
		public double RestY
		{
			get { return _restY; }
			set
			{
				if (_restY == value)
					return;

				_restY = value;

				_translateTransformRest.OffsetY = _restY;
			}
		}

		private double _restZ;
		/// <summary>
		/// Gets or sets this Bone's resting Z coordinate.
		/// </summary>
		public double RestZ
		{
			get { return _restZ; }
			set
			{
				if (_restZ == value)
					return;

				_restZ = value;

				_translateTransformRest.OffsetZ = _restZ;
			}
		}

		private double _restAngleX;
		/// <summary>
		/// Gets or sets this Bone's resting X-axis angle (roll) in degrees
		/// (Rotation convention is yaw-pitch-roll / YZX.)
		/// </summary>
		public double RestAngleX
		{
			get { return _restAngleX; }
			set
			{
				if (_restAngleX == value)
					return;

				_restAngleX = value;

				((AxisAngleRotation3D)_rotateTransformXRest.Rotation).Angle = _restAngleX;
			}
		}

		private double _restAngleY;
		/// <summary>
		/// Gets or sets this Bone's resting Y-axis angle (yaw) in degrees
		/// (Rotation convention is yaw-pitch-roll / YZX.)
		/// </summary>
		public double RestAngleY
		{
			get { return _restAngleY; }
			set
			{
				if (_restAngleY == value)
					return;

				_restAngleY = value;

				((AxisAngleRotation3D)_rotateTransformYRest.Rotation).Angle = _restAngleY;
			}
		}

		private double _restAngleZ;
		/// <summary>
		/// Gets or sets this Bone's resting Z-axis angle (pitch) in degrees
		/// (Rotation convention is yaw-pitch-roll / YZX.)
		/// </summary>
		public double RestAngleZ
		{
			get { return _restAngleZ; }
			set
			{
				if (_restAngleZ == value)
					return;

				_restAngleZ = value;

				((AxisAngleRotation3D)_rotateTransformZRest.Rotation).Angle = _restAngleZ;
			}
		}

		private double _restScaleX;
		/// <summary>
		/// Gets or sets this Bone's resting X-axis scale factor.
		/// </summary>
		public double RestScaleX
		{
			get { return _restScaleX; }
			set
			{
				if (_restScaleX == value)
					return;

				_restScaleX = value;

				_scaleTransformRest.ScaleX = _restScaleX;
			}
		}

		private double _restScaleY;
		/// <summary>
		/// Gets or sets this Bone's resting Y-axis scale factor.
		/// </summary>
		public double RestScaleY
		{
			get { return _restScaleY; }
			set
			{
				if (_restScaleY == value)
					return;

				_restScaleY = value;

				_scaleTransformRest.ScaleY = _restScaleY;
			}
		}

		private double _restScaleZ;
		/// <summary>
		/// Gets or sets this Bone's resting Z-axis scale factor.
		/// </summary>
		public double RestScaleZ
		{
			get { return _restScaleZ; }
			set
			{
				if (_restScaleZ == value)
					return;

				_restScaleZ = value;

				_scaleTransformRest.ScaleZ = _restScaleZ;
			}
		}
		#endregion

		#region Public Methods
		/// <summary>
		/// Creates a deep copy of this Bone.
		/// </summary>
		/// <returns></returns>
		public override object Clone()
		{
			Bone clone = new Bone
			{
				Name = Name,
				Parent = Parent,

				RestX = RestX,
				RestY = RestY,
				RestZ = RestZ,
				RestAngleX = RestAngleX,
				RestAngleY = RestAngleY,
				RestAngleZ = RestAngleZ,
				ScaleX = ScaleX,
				ScaleY = ScaleY,
				ScaleZ = ScaleZ
			};

			ClonePose(clone);

			for (int i = 0; i < VertexInfluences.Count(); i++)
				clone.VertexInfluences.Add(VertexInfluences[i]);

			for (int c = 0; c < Children.Count(); c++)
				clone.Children.Add(Children[c].Clone() as Bone);

			for (int a = 0; a < RigidAttachments.Count(); a++)
				clone.RigidAttachments.Add(RigidAttachments[a].Clone() as SceneNode);

			return clone;
		}
		#endregion

		#region Internal Methods
		internal void Rest()
		{
			X = RestX;
			Y = RestY;
			Z = RestZ;
			AngleX = RestAngleX;
			AngleY = RestAngleY;
			AngleZ = RestAngleZ;
			ScaleX = RestScaleX;
			ScaleY = RestScaleY;
			ScaleZ = RestScaleZ;
		}

		internal Transform3D ComputeSkinTransform()
		{
			Transform3DGroup trans = new Transform3DGroup();
			trans.Children.Add((Transform3D)GetRestingTransformRelativeToSkin().Inverse);
			trans.Children.Add(GetCurrentTransformRelativeToSkin());

			return trans;
		}

		static internal Point3D TransformSkinPoint(Transform3D skinTransform, Point3D point)
		{
			return skinTransform.Transform(point);
		}

		static internal Vector3D TransformSkinNormal(Transform3D skinTransform, Vector3D normal)
		{
			// Since vectors do not have a location, they only need to be transformed
			// by the rotation portion of the skin transform.
			Matrix3D mat = skinTransform.Value;
			mat.OffsetX = 0;
			mat.OffsetY = 0;
			mat.OffsetZ = 0;

			MatrixTransform3D skinRotationTransform = new MatrixTransform3D(mat);

			return skinRotationTransform.Transform(normal);
		}
		#endregion

		#region Private Methods
		private void SetupRestTransforms()
		{
			// Setup resting translation and scale transforms
			_translateTransformRest = new TranslateTransform3D();
			_scaleTransformRest = new ScaleTransform3D();

			// Setup resting rotation transform using YZX Euler angles
			Transform3DGroup rotateTransformRest = new Transform3DGroup();
			_rotateTransformXRest = new RotateTransform3D(new AxisAngleRotation3D(new Vector3D(1, 0, 0), 0));
			_rotateTransformYRest = new RotateTransform3D(new AxisAngleRotation3D(new Vector3D(0, 1, 0), 0));
			_rotateTransformZRest = new RotateTransform3D(new AxisAngleRotation3D(new Vector3D(0, 0, 1), 0));
			rotateTransformRest.Children.Add(_rotateTransformYRest); // => yaw
			rotateTransformRest.Children.Add(_rotateTransformZRest); // => pitch
			rotateTransformRest.Children.Add(_rotateTransformXRest); // => roll

			// Setup total resting transform (order of children matters)
			_restingTransform = new Transform3DGroup();
			_restingTransform.Children.Add(_scaleTransformRest);
			_restingTransform.Children.Add(rotateTransformRest);
			_restingTransform.Children.Add(_translateTransformRest);
		}

		private Transform3D GetCurrentTransformRelativeToSkin()
		{
			Transform3DGroup ret = new Transform3DGroup();
			ret.Children.Add(Data.Transform);

			if (Parent != null)
				ret.Children.Add(Parent.GetCurrentTransformRelativeToSkin());

			return ret;
		}

		private Transform3D GetRestingTransformRelativeToSkin()
		{
			Transform3DGroup ret = new Transform3DGroup();
			ret.Children.Add(_restingTransform);

			if (Parent != null)
				ret.Children.Add(Parent.GetCurrentTransformRelativeToSkin());

			return ret;
		}

		private void UpdateArmature()
		{
			if (Armature != null)
				Armature.UpdateBoneTransforms();
		}
		#endregion

		#region Event Handlers
		private void Children_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
		{
			switch (e.Action)
			{
				case NotifyCollectionChangedAction.Add:
					HandleCollectionActionAdd(sender, e);
					break;

				case NotifyCollectionChangedAction.Remove:
					HandleCollectionActionRemove(sender, e);
					break;

				default:
					throw new NotSupportedException("NotifyCollectionChangedAction.Reset is not supported.");
			}

			UpdateArmature();
		}

		private void HandleCollectionActionAdd(object sender, NotifyCollectionChangedEventArgs e)
		{
			if (sender is SceneNodeCollection)
			{
				foreach (SceneNode node in e.NewItems)
				{
					// Clear attachment to current parent
					node.Parent = null;

					// Add data to model group
					node.BoneParent = this;
					AttachedVisualChildren.Add(node.Data);
				}

				return;
			}

			foreach (Bone child in e.NewItems)
			{
				// Add child's data to child model group
				VisualChildren.Add(child.Data);

				// Remove child from any current parent
				if (child.Parent != null)
					child.Parent.Children.Remove(child);  // Will set child.Parent = null

				// Set parent
				child._parent = this;
			}
		}

		private void HandleCollectionActionRemove(object sender, NotifyCollectionChangedEventArgs e)
		{
			if (sender is SceneNodeCollection)
			{
				foreach (SceneNode node in e.OldItems)
				{
					// Remove data from model group
					node.BoneParent = null;
					AttachedVisualChildren.Remove(node.Data);
				}

				return;
			}

			foreach (Bone child in e.OldItems)
			{
				// Remove child's data from model group
				VisualChildren.Remove(child.Data);

				// Clear parent
				child._parent = null;
			}
		}

		void VertexInfluences_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
		{
			// Since influences have changed, update the Armature's set of weighted transforms
			UpdateArmature();
		}
		#endregion
	}
}
