﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Media.Media3D;
using System.Linq;
using System.Xml.Linq;
using System.Reflection;
using System.Drawing;
using System.Windows.Interop;
using System.Runtime.InteropServices;

namespace Ascend
{
	/// <summary>
	/// Helper class for loading Ascend (*.asc) models.
	/// </summary>
	internal static class Loader
	{
		#region Constants
		private const double RAD_TO_DEG = 180 / Math.PI;
		#endregion

		#region Private Members
		private static string WorkingDirectory;
		private static Dictionary<string, Bitmap> Textures;
		private static Dictionary<string, Stream> TextureStreams;
		#endregion

		#region Public Methods
		public static ModelSceneNode Load(string filename)
		{
			WorkingDirectory = Path.GetDirectoryName(filename);
			Textures = null;
			TextureStreams = null;

			try
			{
				XDocument doc = XDocument.Load(filename);
				return ReadSceneNode((XElement)doc.Element("Ascend").FirstNode);
			}
			catch (Exception ex)
			{
				throw new Exception("Unable to load Ascend model \'" + filename + "\'. See the InnerException property for more information.", ex);
			}
		}

		public static ModelSceneNode Load(Stream stream)
		{
			try
			{
				WorkingDirectory = null;
				TextureStreams = null;
				Textures = null;
				XDocument doc = XDocument.Load(stream);

				return ReadSceneNode((XElement)doc.Element("Ascend").FirstNode);
			}
			catch (Exception ex)
			{
				throw new Exception("Unable to load Ascend model from stream. See the InnerException property for more information.", ex);
			}
		}

		public static ModelSceneNode Load(Stream stream, Dictionary<string, Bitmap> textures)
		{
			try
			{
				WorkingDirectory = null;
				TextureStreams = null;
				Textures = textures;
				XDocument doc = XDocument.Load(stream);

				return ReadSceneNode((XElement)doc.Element("Ascend").FirstNode);
			}
			catch (Exception ex)
			{
				throw new Exception("Unable to load Ascend model from stream. See the InnerException property for more information.", ex);
			}
		}

		public static ModelSceneNode Load(Stream stream, Dictionary<string, Stream> textureStreams)
		{
			try
			{
				WorkingDirectory = null;
				Textures = null;
				TextureStreams = textureStreams;
				XDocument doc = XDocument.Load(stream);

				return ReadSceneNode((XElement)doc.Element("Ascend").FirstNode);
			}
			catch (Exception ex)
			{
				throw new Exception("Unable to load Ascend model from stream. See the InnerException property for more information.", ex);
			}
		}
		#endregion

		#region Private Methods
		private static ModelSceneNode ReadSceneNode(XElement xml)
		{
			ModelSceneNode node = null;

			// Get SceneNode type
			switch (xml.Name.LocalName)
			{
				case "ModelSceneNode":
					node = new ModelSceneNode();
					break;
				case "AnimatedSceneNode":
					node = new AnimatedSceneNode();
					break;
				default:
					throw new InvalidDataException("Model file does not contain a valid SceneNode.");
			}

			// Read name
			node.Name = xml.Attribute("Name").Value;

			// Read transform
			ReadTransform(node, xml);

			// Read geometry
			List<List<int>> vertexAliases = ReadGeometry(node, xml);

			// Read color/texture
			ReadMaterial(node, xml);

			if (node is AnimatedSceneNode)
			{
				// Read animations
				ReadSceneNodeAnimationFrames((AnimatedSceneNode)node, xml);

				// Read armature
				ReadArmature((AnimatedSceneNode)node, vertexAliases, xml);

				// Read animation definitions
				ReadAnimationDefinitions((AnimatedSceneNode)node, xml);
			}

			// Read children
			foreach (var xmlChild in xml.Element("Children").Elements())
				node.Children.Add(ReadSceneNode(xmlChild));

			return node;
		}

		private static void ReadTransform(TransformableObject node, XElement xml)
		{
			ObjectPose pose = ReadObjectPose(xml);

			node.X = pose.X;
			node.Y = pose.Y;
			node.Z = pose.Z;

			node.AngleX = pose.AngleX;
			node.AngleY = pose.AngleY;
			node.AngleZ = pose.AngleZ;

			node.ScaleX = pose.ScaleX;
			node.ScaleY = pose.ScaleY;
			node.ScaleZ = pose.ScaleZ;
		}

		private static void ReadRestTransform(Bone bone, XElement xml)
		{
			ObjectPose pose = ReadObjectPose(xml);

			bone.RestX = pose.X;
			bone.RestY = pose.Y;
			bone.RestZ = pose.Z;

			bone.RestAngleX = pose.AngleX;
			bone.RestAngleY = pose.AngleY;
			bone.RestAngleZ = pose.AngleZ;

			bone.RestScaleX = pose.ScaleX;
			bone.RestScaleY = pose.ScaleY;
			bone.RestScaleZ = pose.ScaleZ;
		}

		private static ObjectPose ReadObjectPose(XElement xml)
		{
			double[] pos = xml.Attribute("Position").Value.Split().Select(val => double.Parse(val)).ToArray();
			double[] rot = xml.Attribute("Rotation").Value.Split().Select(val => double.Parse(val)).ToArray();
			double[] scl = xml.Attribute("Scale").Value.Split().Select(val => double.Parse(val)).ToArray();

			return new ObjectPose
			{
				X = pos[0],
				Y = pos[1],
				Z = pos[2],

				AngleX = rot[0] * RAD_TO_DEG,
				AngleY = rot[1] * RAD_TO_DEG,
				AngleZ = rot[2] * RAD_TO_DEG,

				ScaleX = scl[0],
				ScaleY = scl[1],
				ScaleZ = scl[2],
			};
		}

		private static void ReadMaterial(ModelSceneNode node, XElement xml)
		{
			if (xml.Attribute("Texture") != null && TryLoadTexture(node, xml.Attribute("Texture").Value))
				return;

			// If texture loading failed, set color
			float[] col = xml.Attribute("Color").Value.Split().Select(val => float.Parse(val)).ToArray();

			node.Model.Material = new DiffuseMaterial(new SolidColorBrush(
				new System.Windows.Media.Color
				{
					ScR = col[0],
					ScG = col[1],
					ScB = col[2],
					ScA = col[3]
				}
			));
		}

		private static bool TryLoadTexture(ModelSceneNode node, string texture)
		{
			ImageSource img = null;

			// Load texture from file path, bitmap resource, or stream
			if (WorkingDirectory != null)
			{
				img = LoadTextureFromPath(texture);
			}
			else if (Textures != null && Textures.ContainsKey(texture))
			{
				img = LoadTextureFromBitmap(Textures[texture]);
			}
			else if (TextureStreams != null && TextureStreams.ContainsKey(texture))
			{
				img = LoadTextureFromStream(TextureStreams[texture]);
			}

			if (img != null)
			{
				node.Model.Material = new DiffuseMaterial(new ImageBrush(img)
				{
					TileMode = TileMode.Tile,
					ViewportUnits = BrushMappingMode.Absolute,
				});

				return true;
			}

			return false;
		}

		private static BitmapSource LoadTextureFromPath(string texture)
		{
			string texturePath = Path.Combine(WorkingDirectory, texture);

			if (!File.Exists(texturePath))
				return null;

			return new BitmapImage(new Uri(texturePath, UriKind.Relative));
		}

		[DllImport("gdi32.dll")]
		private static extern bool DeleteObject(IntPtr hObject);

		private static BitmapSource LoadTextureFromBitmap(Bitmap bitmap)
		{
			BitmapSource src = null;
			IntPtr hBitmap = bitmap.GetHbitmap();

			try
			{
				src = Imaging.CreateBitmapSourceFromHBitmap(hBitmap, IntPtr.Zero, Int32Rect.Empty, BitmapSizeOptions.FromEmptyOptions());
			}
			finally
			{
				DeleteObject(hBitmap);
			}

			return src;
		}

		private static BitmapSource LoadTextureFromStream(Stream stream)
		{
			BitmapImage img = new BitmapImage();
			img.BeginInit();
			img.CacheOption = BitmapCacheOption.OnLoad;
			img.StreamSource = stream;
			img.EndInit();

			return img;
		}

		private static List<List<int>> ReadGeometry(ModelSceneNode node, XElement xml)
		{
			List<List<int>> vertexAliases;
			List<Point3D> sharedVertices = ReadVertices(xml, out vertexAliases);
			MeshGeometry3D mesh = new MeshGeometry3D();
			XElement xmlFaces = xml.Element("Faces");

			foreach (var xmlFace in xmlFaces.Elements("Face"))
			{
				// Reference index for computing triangle indices
				int idxRef = mesh.Positions.Count();

				foreach (var xmlFaceVert in xmlFace.Elements("FaceVertex"))
				{
					int i = int.Parse(xmlFaceVert.Attribute("Index").Value);

					// Add vertex to mesh
					vertexAliases[i].Add(mesh.Positions.Count());
					mesh.Positions.Add(sharedVertices[i]);

					// Check for normals
					if (xmlFaceVert.Attribute("Normal") != null)
					{
						double[] norm = xmlFaceVert.Attribute("Normal").Value.Split().Select(val => double.Parse(val)).ToArray();
						mesh.Normals.Add(new Vector3D(norm[0], norm[1], norm[2]));
					}

					// Check for UVs
					if (xmlFaceVert.Attribute("UV") != null)
					{
						double[] uv = xmlFaceVert.Attribute("UV").Value.Split().Select(val => double.Parse(val)).ToArray();
						mesh.TextureCoordinates.Add(new System.Windows.Point(uv[0], uv[1]));
					}
				}

				// Compute triangle indices for face
				int[] deltaIndices = (xmlFace.Elements().Count() == 3)
					? new int[] { 0, 1, 2 }
					: new int[] { 0, 1, 2, 2, 3, 0 };

				foreach (int i in deltaIndices)
					mesh.TriangleIndices.Add(idxRef + i);
			}

			node.Model = new GeometryModel3D { Geometry = mesh };

			return vertexAliases;
		}

		private static List<Point3D> ReadVertices(XElement xml, out List<List<int>> aliases)
		{
			List<Point3D> vertices = new List<Point3D>();
			XElement xmlVertices = xml.Element("Vertices");

			foreach (var xmlVert in xmlVertices.Elements("Vertex"))
			{
				vertices.Add(new Point3D(
					double.Parse(xmlVert.Attribute("X").Value),
					double.Parse(xmlVert.Attribute("Y").Value),
					double.Parse(xmlVert.Attribute("Z").Value)
				));
			}

			// Initialize vertex aliases
			aliases = new List<List<int>>();

			for (int i = 0; i < vertices.Count(); i++)
				aliases.Add(new List<int>());

			return vertices;
		}

		private static void ReadSceneNodeAnimationFrames(AnimatedSceneNode node, XElement xml)
		{
			if (xml.Element("Frames") == null)
				return;

			foreach (var xmlFrame in xml.Element("Frames").Elements("Frame"))
			{
				node.AnimationFrames.Add(new AnimationFrame
				{
					ModelPose = ReadObjectPose(xmlFrame)
				}
				);
			}
		}

		private static void ReadBoneAnimationFrames(Bone bone, Armature arm, XElement xml)
		{
			if (xml.Element("Frames") == null)
				return;

			int frameNumber = 0;
			foreach (var xmlFrame in xml.Element("Frames").Elements("Frame"))
			{
				if (!SkinHasFrame(arm.Skin, frameNumber))
				{
					// If skin does not have the frame at this point,
					// there are no rigid animations for it; this frame
					// will only include bone poses.
					arm.Skin.AnimationFrames.Add(new AnimationFrame());
				}

				// Set the bone pose for this frame
				arm.Skin.AnimationFrames[frameNumber].BonePoses.Add(ReadObjectPose(xmlFrame));

				frameNumber++;
			}
		}

		private static bool SkinHasFrame(AnimatedSceneNode skin, int frameNumber)
		{
			return frameNumber < skin.AnimationFrames.Count();
		}

		private static void ReadAnimationDefinitions(AnimatedSceneNode node, XElement xml)
		{
			if (xml.Element("Animations") == null)
				return;

			foreach (var xmlAnim in xml.Element("Animations").Elements("Animation"))
			{
				node.DefineAnimation(xmlAnim.Attribute("Name").Value,
					new FrameInterval(
						int.Parse(xmlAnim.Attribute("FrameStart").Value),
						int.Parse(xmlAnim.Attribute("FrameEnd").Value)
					)
				);
			}
		}

		private static void ReadArmature(AnimatedSceneNode node, List<List<int>> vertexAliases, XElement xml)
		{
			XElement xmlArm = xml.Element("Armature");

			if (xmlArm == null)
				return;

			node.Armature = new Armature();

			// Read name
			node.Armature.Name = xmlArm.Attribute("Name").Value;

			// Read root Bones
			foreach (var xmlBone in xmlArm.Element("RootBones").Elements())
				node.Armature.RootBones.Add(ReadBone(node.Armature, vertexAliases, xmlBone));
		}

		private static Bone ReadBone(Armature arm, List<List<int>> vertexAliases, XElement xml)
		{
			Bone bone = new Bone();

			// Read name
			bone.Name = xml.Attribute("Name").Value;

			// Read rest transform
			ReadRestTransform(bone, xml);
			bone.Rest();

			// Read vertex influences
			ReadBoneInfluences(bone, vertexAliases, xml);

			// Read animation frames
			ReadBoneAnimationFrames(bone, arm, xml);

			// Read children
			foreach (var xmlChild in xml.Element("Children").Elements())
			{
				if (xmlChild.Name == "Bone")
					bone.Children.Add(ReadBone(arm, vertexAliases, xmlChild));
				else
					bone.RigidAttachments.Add(ReadSceneNode(xmlChild));
			}

			return bone;
		}

		private static void ReadBoneInfluences(Bone bone, List<List<int>> vertexAliases, XElement xml)
		{
			XElement xmlInfluences = xml.Element("Influences");

			foreach (var xmlInf in xmlInfluences.Elements("Influence"))
			{
				VertexInfluence inf = new VertexInfluence(
					double.Parse(xmlInf.Attribute("Weight").Value),
					vertexAliases[int.Parse(xmlInf.Attribute("Vertex").Value)]
				);

				bone.VertexInfluences.Add(inf);
			}

		}
		#endregion
	}
}