﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Windows.Media;
using System.Windows.Media.Media3D;

namespace Ascend
{
	/// <summary>
	/// Class for generating basic geometric models.
	/// </summary>
	public static class ModelMaker
	{
		#region Public Methods
		/// <summary>
		/// Generates a triangle from the given points.
		/// </summary>
		/// <param name="p0">The first point in the triangle.</param>
		/// <param name="p1">The second point in the triangle.</param>
		/// <param name="p2">The third point in the triangle.</param>
		/// <param name="color">The color of the triangle.</param>
		/// <param name="doubleSided">Whether or not the triangle should be double sided.</param>
		public static GeometryModel3D MakeTriangle(Point3D p0, Point3D p1, Point3D p2, Color color, bool doubleSided = false)
		{
			MeshGeometry3D mesh = new MeshGeometry3D();

			mesh.Positions.Add(p0);
			mesh.Positions.Add(p1);
			mesh.Positions.Add(p2);

			mesh.TriangleIndices.Add(0);
			mesh.TriangleIndices.Add(1);
			mesh.TriangleIndices.Add(2);

			Material material = new DiffuseMaterial(new SolidColorBrush(color));

			return new GeometryModel3D(mesh, material) { BackMaterial = doubleSided ? material : null };
		}

		/// <summary>
		/// Generates a unit cube.
		/// </summary>
		/// <param name="color">The color of the cube.</param>
		public static GeometryModel3D MakeCube(Color color)
		{
			Point3D[] vertices =
			{
				new Point3D(-0.5, 0.5, -0.5),
				new Point3D(0.5, 0.5, -0.5),
				new Point3D(0.5, 0.5, 0.5),
				new Point3D(-0.5, 0.5, 0.5),
				new Point3D(-0.5, -0.5, -0.5),
				new Point3D(0.5, -0.5, -0.5),
				new Point3D(0.5, -0.5, 0.5),
				new Point3D(-0.5, -0.5, 0.5)
			};

			MeshGeometry3D mesh = new MeshGeometry3D();

			int[][] indices =
			{
				new int[] {3, 2, 1, 0},
				new int[] {6, 7, 4, 5},
				new int[] {7, 6, 2, 3},
				new int[] {5, 4, 0, 1},
				new int[] {4, 7, 3, 0},
				new int[] {6, 5, 1, 2}
			};

			for (int f = 0; f < 6; f++)
			{
				int[] verts = indices[f];

				int idxRef = mesh.Positions.Count;

				mesh.Positions.Add(vertices[verts[0]]);
				mesh.Positions.Add(vertices[verts[1]]);
				mesh.Positions.Add(vertices[verts[2]]);
				mesh.Positions.Add(vertices[verts[3]]);

				mesh.TriangleIndices.Add(idxRef);
				mesh.TriangleIndices.Add(idxRef + 1);
				mesh.TriangleIndices.Add(idxRef + 2);
				mesh.TriangleIndices.Add(idxRef + 2);
				mesh.TriangleIndices.Add(idxRef + 3);
				mesh.TriangleIndices.Add(idxRef);
			}

			Material material = new DiffuseMaterial(new SolidColorBrush(color));

			return new GeometryModel3D(mesh, material);
		}

		/// <summary>
		/// Generates a 0.5 radius sphere of specified resolution.
		/// </summary>
		/// <param name="horizontalSegments">Horizontal resolution.</param>
		/// <param name="verticalSegments">Vertical resolution.</param>
		/// <param name="color">The color of the sphere.</param>
		public static GeometryModel3D MakeSphere(int horizontalSegments, int verticalSegments, Color color)
		{
			MeshGeometry3D mesh = new MeshGeometry3D();

			double refAlpha = Math.PI / 2;
			double deltaAlpha = Math.PI / verticalSegments;
			double deltaTheta = 2 * Math.PI / horizontalSegments;
			int lastIndex = horizontalSegments * (verticalSegments - 1) + 1;

			mesh.Positions.Add(new Point3D(0, 0.5, 0));

			for (int r = 1; r < verticalSegments; r++)
			{
				double alpha = refAlpha - r * deltaAlpha;
				double y = 0.5 * Math.Sin(alpha);
				double currRad = 0.5 * Math.Cos(alpha);

				for (int c = 0; c < horizontalSegments; c++)
				{
					double theta = c * deltaTheta;

					mesh.Positions.Add(new Point3D(currRad * Math.Cos(theta), y, currRad * Math.Sin(theta)));

					int i0 = 1 + (r - 1) * horizontalSegments + c;
					int i1 = (c == horizontalSegments - 1) ? i0 - horizontalSegments + 1 : i0 + 1;
					int i2 = i1 + horizontalSegments;
					int i3 = i0 + horizontalSegments;

					if (r < verticalSegments - 1)
					{
						mesh.TriangleIndices.Add(i0);
						mesh.TriangleIndices.Add(i1);
						mesh.TriangleIndices.Add(i2);

						mesh.TriangleIndices.Add(i2);
						mesh.TriangleIndices.Add(i3);
						mesh.TriangleIndices.Add(i0);
					}

					if (r == 1)
					{
						mesh.TriangleIndices.Add(0);
						mesh.TriangleIndices.Add(i1);
						mesh.TriangleIndices.Add(i0);

					}

					if (r == verticalSegments - 1)
					{
						mesh.TriangleIndices.Add(i0);
						mesh.TriangleIndices.Add(i1);
						mesh.TriangleIndices.Add(lastIndex);
					}
				}
			}

			mesh.Positions.Add(new Point3D(0, -0.5, 0));

			Material material = new DiffuseMaterial(new SolidColorBrush(color));

			return new GeometryModel3D(mesh, material);
		}

		/// <summary>
		/// Generates a 0.5 radius sphere of 16-by-8 resolution.
		/// </summary>
		/// <param name="color">The color of the sphere.</param>
		public static GeometryModel3D MakeSphere(Color color)
		{
			return MakeSphere(16, 8, color);
		}
		#endregion

		#region Private Methods
		private static double ToDegrees(double radians)
		{
			return radians * 180 / Math.PI;
		}

		private static double ToRadians(double degrees)
		{
			return degrees * Math.PI / 180;
		}
		#endregion
	}
}
