﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System.Collections.Generic;
using System.Windows;
using System.Windows.Media.Media3D;

namespace Ascend
{
	/// <summary>
	/// Contains a root SceneNode and a camera and is viewable via SceneViewer3D.
	/// </summary>
	public class Scene3D : DependencyObject
	{
		#region Lifecycle
		/// <summary>
		/// Creates an empty Scene3D instance.
		/// </summary>
		public Scene3D()
		{
			// Set default render root SceneNode
			RenderRoot = new ModelSceneNode();
			RenderRoot.Name = "__root__";

			// Set default camera
			Camera = new PerspectiveCamera(new Point3D(0, 0, 1), new Vector3D(0, 0, -1), new Vector3D(0, 1, 0), 70.0);
		}
		#endregion

		#region Dependency Properties
		/// <summary>
		/// Gets the root of this Scene3D's visual graph.
		/// </summary>
		public ModelSceneNode RenderRoot
		{
			get { return (ModelSceneNode)GetValue(RenderRootProperty); }
			private set { SetValue(RenderRootProperty, value); }
		}

		public static readonly DependencyProperty RenderRootProperty =
			DependencyProperty.Register("RenderRoot", typeof(ModelSceneNode), typeof(Scene3D));


		/// <summary>
		/// Gets or sets the primary camera for this Scene3D.
		/// </summary>
		public ProjectionCamera Camera
		{
			get { return (ProjectionCamera)GetValue(CameraProperty); }
			set { SetValue(CameraProperty, value); }
		}

		public static readonly DependencyProperty CameraProperty =
			DependencyProperty.Register("Camera", typeof(ProjectionCamera), typeof(Scene3D));
		#endregion

		#region Public Methods
		/// <summary>
		/// Points the camera to the specified point.
		/// </summary>
		/// <param name="point">The point at which to look.</param>
		public void LookAt(Point3D point)
		{
			if (Camera != null)
			{
				Camera.LookDirection = point - Camera.Position;
				Camera.UpDirection = new Vector3D(0, Camera.UpDirection.Y, 0);  // Keep from rolling the camera
			}
		}

		/// <summary>
		/// Points the camera to the specified TransformableObject.
		/// </summary>
		/// <param name="target">The TransformableObject at which to look.</param>
		public void LookAt(TransformableObject target)
		{
			LookAt(target.GlobalTransform.Transform(new Point3D(0, 0, 0)));
		}
		#endregion
	}
}
