﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Media.Media3D;
using System.Windows.Threading;

namespace Ascend
{
	/// <summary>
	/// An enumeration of animation states for AnimatedSceneNodes.
	/// </summary>
	public enum AnimationState
	{
		Stopped,
		Playing,
		PlayingInTimeline
	}


	/// <summary>
	/// A ModelSceneNode that contains animations.
	/// </summary>
	public class AnimatedSceneNode : ModelSceneNode
	{
		#region Lifecycle
		/// <summary>
		/// Creates a new empty AnimatedSceneNode instance.
		/// </summary>
		public AnimatedSceneNode()
		{
			// Create the default armature content
			Data.Children.Add(new Model3DGroup());

			AnimationFrames = new List<AnimationFrame>();

			_animations = new Dictionary<string, FrameInterval>();
		}

		/// <summary>
		/// Creates a new AnimatedSceneNode instance with given model.
		/// </summary>
		/// <param name="model">The model for the AnimatedSceneNode.</param>
		public AnimatedSceneNode(GeometryModel3D model)
			: this()
		{
			Model = model;
		}
		#endregion

		#region Events
		/// <summary>
		/// Occurs when a non-looping animation completes.
		/// </summary>
		public event SceneNodeAnimationEventHandler AnimationCompleted;
		#endregion

		#region Private Members
		private object _animLock = new object();
		private Dictionary<string, FrameInterval> _animations;

		private Model3DGroup ArmatureVisual
		{
			get { return (Model3DGroup)Data.Children[2]; }
			set { Data.Children[2] = value; }
		}
		#endregion

		#region Properties
		/// <summary>
		/// Gets a read-only list of TransformableObjects that are transformed relative to this AnimatedSceneNode.
		/// </summary>
		public override IList<TransformableObject> TransformChildren
		{
			get
			{
				IEnumerable<TransformableObject> transChildren;

				if (Armature != null)
					transChildren = Armature.RootBones.Cast<TransformableObject>().Concat(Children.Cast<TransformableObject>());
				else
					transChildren = Children.Cast<TransformableObject>();

				return transChildren.ToList().AsReadOnly();
			}
		}

		/// <summary>
		/// Gets this AnimatedSceneNode's list of animation frames.
		/// </summary>
		public List<AnimationFrame> AnimationFrames { get; private set; }

		private AnimationFrame _currentFrame;
		/// <summary>
		/// Gets or sets this AnimatedSceneNode's current animation frame.
		/// </summary>
		public AnimationFrame CurrentFrame
		{
			get { return _currentFrame; }
			set
			{
				if (_currentFrame == value)
					return;

				if (value != null && !AnimationFrames.Contains(value))
					throw new ArgumentException("CurrentFrame must be a value from AnimationFrames.");

				_currentFrame = value;

				if (_currentFrame == null)
					return;

				if (_currentFrame.ModelPose != null)
				{
					X = _currentFrame.ModelPose.X;
					Y = _currentFrame.ModelPose.Y;
					Z = _currentFrame.ModelPose.Z;

					AngleX = _currentFrame.ModelPose.AngleX;
					AngleY = _currentFrame.ModelPose.AngleY;
					AngleZ = _currentFrame.ModelPose.AngleZ;

					ScaleX = _currentFrame.ModelPose.ScaleX;
					ScaleY = _currentFrame.ModelPose.ScaleY;
					ScaleZ = _currentFrame.ModelPose.ScaleZ;
				}

				if (Armature != null && _currentFrame.BonePoses.Count > 0)
					Armature.ApplyPoses(_currentFrame.BonePoses);
			}
		}

		/// <summary>
		/// Get this AnimatedSceneNode's current animation state.
		/// </summary>
		public AnimationState AnimationState { get; internal set; }

		/// <summary>
		/// Gets the name of the currently playing animation (may be null).
		/// </summary>
		public string CurrentAnimation { get; internal set; }

		private Armature _armature;
		/// <summary>
		/// Gets or sets this AnimatedSceneNode's Armature.
		/// </summary>
		public Armature Armature
		{
			get { return _armature; }
			set
			{
				if (_armature == value)
					return;

				// Clear old visual
				ArmatureVisual = new Model3DGroup();

				_armature = value;

				if (_armature != null)
				{
					_armature.Skin = this;
					ArmatureVisual = _armature.Data;
				}
			}
		}
		#endregion

		#region Public Methods
		/// <summary>
		/// Play the AnimatedSceneNode's specified animation.
		/// </summary>
		/// <param name="animationName">The name of the animation frame interval to play.</param>
		/// <param name="loop">Whether or not to loop the animation.</param>
		/// <param name="restart">Whether or not to start the animation from the beginning.</param>
		/// <param name="fps">Animation speed in frames per second. Specifying a negative value
		/// will play animation in reverse.</param>
        /// <exception cref="ArgumentException">
        /// Thrown if <paramref name="animationName"/> is not a defined animation or
        /// if <paramref name="fps"/> is zero.
        /// </exception>
        /// <exception cref="InvalidOperationException">
        /// Thrown if the AnimatedSceneNode is currently being animated in a Timeline.
        /// </exception>
		public void Play(string animationName, bool loop = false, bool restart = false, double fps = 60.0)
		{
			if (!_animations.ContainsKey(animationName))
				throw new ArgumentException(string.Format("Animation '{0}' does not exist on AnimatedSceneNode.", animationName));

			if (AnimationState == AnimationState.PlayingInTimeline)
				throw new InvalidOperationException("Cannot play an AnimatedSceneNode while it is being animated in a Timeline.");

			if (AnimationState == AnimationState.Playing)
				Pause();

            if (fps == 0.0)
                throw new ArgumentException("Frame rate must be a non-zero value.");

			FrameInterval interval = _animations[animationName];

			List<AnimationFrame> frames = AnimationFrames.GetRange(interval.Start, interval.Length);

			bool reverse = fps < 0.0;

			if (restart)
				CurrentFrame = null;  // Will cause animation to restart (see beginning of ProgressAnimation method)

			StartAnimationThread(animationName, loop, reverse, fps, frames);
		}

		/// <summary>
		/// Pauses any current animation.
		/// </summary>
		public void Pause()
		{
			// This should stop the animation thread
			AnimationState = AnimationState.Stopped;

			CurrentAnimation = null;
		}

		/// <summary>
		/// Defines an animation frame interval on this AnimatedSceneNode. If a definition already exists
		/// by the given name, it will be overwritten.
		/// </summary>
		/// <param name="animationName">Name of the animation.</param>
		/// <param name="frameInterval">Frame interval that defines the animation.</param>
        /// <exception cref="ArgumentOutOfRangeException">
        /// Thrown if <paramref name="frameInterval"/> is outside of the range of AnimationFrames.
        /// </exception>
		public void DefineAnimation(string animationName, FrameInterval frameInterval)
		{
			if (frameInterval.Start < 0 || frameInterval.End >= AnimationFrames.Count)
				throw new ArgumentOutOfRangeException("Interval outside index range of AnimationFrames");

			if (_animations.ContainsKey(animationName))
				_animations[animationName] = frameInterval;
			else
				_animations.Add(animationName, frameInterval);
		}

		/// <summary>
		/// Defines an animation frame interval on this AnimatedSceneNode. If a definition already exists
		/// by the given name, it will be overwritten.
		/// </summary>
		/// <param name="animationName">Name of the animation.</param>
		/// <param name="start">Starting animation frame index.</param>
		/// <param name="end">Ending animation frame index.</param>
        /// <exception cref="ArgumentException">
        /// Thrown if <paramref name="start"/> is greater than <paramref name="end"/>.
        /// </exception>
        /// <exception cref="ArgumentOutOfRangeException">
        /// Thrown if <paramref name="start"/> and <paramref name="end"/> define an interval
        /// outside of the range of AnimationFrames.
        /// </exception>
		public void DefineAnimation(string animationName, int start, int end)
		{
			DefineAnimation(animationName, new FrameInterval(start, end));
		}

		/// <summary>
		/// Removes the specified animation from this AnimatedSceneNode.
		/// </summary>
		/// <param name="animationName">Name of the animation frame interval to remove.</param>
		public void RemoveAnimation(string animationName)
		{
			_animations.Remove(animationName);
		}

		/// <summary>
		/// Returns the specified animation's FrameInterval.
		/// </summary>
		/// <param name="animationName">Name of the animation.</param>
        /// <exception cref="ArgumentException">
        /// Thrown if <paramref name="animationName"/> is not a defined animation.
        /// </exception>
		public FrameInterval GetAnimationInterval(string animationName)
		{
			if (!_animations.ContainsKey(animationName))
				throw new ArgumentException(string.Format("Animation '{0}' does not exist on AnimatedSceneNode.", animationName));

			return _animations[animationName];
		}

		/// <summary>
		/// Returns a list of animations defined on this AnimatedSceneNode.
		/// </summary>
		public List<string> ListAnimations()
		{
			return _animations.Keys.ToList();
		}

		/// <summary>
		/// Creates a deep copy of this AnimatedSceneNode.
		/// </summary>
		public override object Clone()
		{
			AnimatedSceneNode clone = new AnimatedSceneNode(Model.Clone());

			CloneProperties(clone);

			CloneChildren(clone);

			clone.Armature = (Armature != null) ? Armature.Clone() as Armature : null;

			return clone;
		}
		#endregion

		#region Protected Methods
		protected override void CloneProperties(SceneNode clone)
		{
			AnimatedSceneNode animClone = clone as AnimatedSceneNode;

			base.CloneProperties(animClone);

			// Clone animation frames
			foreach (var frame in AnimationFrames)
				animClone.AnimationFrames.Add(frame.Clone() as AnimationFrame);

			// Clone CurrentFrame
			if (CurrentFrame != null)
				animClone.CurrentFrame = animClone.AnimationFrames[AnimationFrames.IndexOf(CurrentFrame)];

			// Clone animation definitions
			foreach (var anim in _animations)
				animClone.DefineAnimation(anim.Key, anim.Value);
		}
		#endregion

		#region Private Methods
		private void StartAnimationThread(string animationName, bool loop, bool reverse, double fps, List<AnimationFrame> frames)
		{
			Task.Factory.StartNew(() =>
				{
					lock (_animLock)
					{
						AnimationState = AnimationState.Playing;
						CurrentAnimation = animationName;

						// Start timing
						DateTime start = DateTime.Now;
						TimeSpan dt;
						int lastFrame = 0;

						while (AnimationState == AnimationState.Playing)
						{
							// Get time and frame deltas
							DateTime thisTime = DateTime.Now;
							dt = thisTime - start;

							int thisFrame = (int)(Math.Abs(fps) * dt.TotalSeconds);
							int df = thisFrame - lastFrame;

							lastFrame = thisFrame;

							if (df >= 1)
							{
								if (ProgressAnimation(df, reverse, loop, frames))
								{
									RaiseAnimationCompletedEvent(animationName);
									break;
								}
							}

							// Sleep to rest CPU
							Thread.Sleep((int)Math.Max(5, 1000 / fps - (DateTime.Now - thisTime).Milliseconds - 5));
						}
					}
				}
			);
		}

		private bool ProgressAnimation(int frameDelta, bool reverse, bool loop, List<AnimationFrame> frames)
		{
			int f = frames.IndexOf(CurrentFrame);

			// Reset frame to beginning or end if current frame not in animation interval
			if (f == -1 && reverse)
				f = frames.Count;

			// Get next frame
			int next = reverse ? f - frameDelta : f + frameDelta;

			// Check for wrap-around
			if ((next <= 0 && reverse) || (next >= frames.Count - 1 && !reverse))
			{
				if (!loop)
				{
					// Clamp to end of animation
					Dispatcher.Invoke(() => CurrentFrame = reverse ? frames.First() : frames.Last());

					Pause();

					// Flag that animation completed
					return true;
				}

				// Compute wrapped-around frame number
				next = (next < 0)
					? (next % frames.Count) + frames.Count
					: next % frames.Count;
			}

			Dispatcher.Invoke(() => CurrentFrame = frames[next]);

			return false;
		}

		private void ResetAnimation(bool reverse, List<AnimationFrame> frames)
		{
			CurrentFrame = reverse ? frames.Last() : frames.First();
		}
		#endregion

		#region Event Invokers
		#region AnimationCompleted
		private void RaiseAnimationCompletedEvent(string animationName)
		{
			if (AnimationCompleted != null)
			{
				SceneNodeAnimationEventArgs e = new SceneNodeAnimationEventArgs { AnimationName = animationName };

				foreach (var del in AnimationCompleted.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion
		#endregion
	}
}