﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Windows.Media.Media3D;

namespace Ascend
{
	/// <summary>
	/// A SceneNode that contains a geometry model.
	/// </summary>
	public class ModelSceneNode : SceneNode
	{
		#region Lifecycle
		/// <summary>
		/// Creates a new default ModelSceneNode instance.
		/// </summary>
		public ModelSceneNode()
		{
			MouseHasEntered = false;

			// Create default model content
			Data.Children.Add(new GeometryModel3D());
		}

		/// <summary>
		/// Creates a new ModelSceneNode instance with given model.
		/// </summary>
		/// <param name="model">The model for the ModelSceneNode</param>
		public ModelSceneNode(GeometryModel3D model)
			: this()
		{
			Model = model;
		}

		/// <summary>
		/// Creates a deep copy of this ModelSceneNode.
		/// </summary>
		public override object Clone()
		{
			ModelSceneNode clone = new ModelSceneNode();
			
			clone.Model = (Model != null) ? Model.Clone() : null;

			CloneProperties(clone);

			CloneChildren(clone);

			return clone;
		}
		#endregion

		#region Events
		/// <summary>
		/// Occurs when a mouse moves onto a ModelSceneNode.
		/// </summary>
		public event SceneNodeMouseEventHandler MouseEnter;

		/// <summary>
		/// Occurs when a mouse moves off of a ModelSceneNode.
		/// </summary>
		public event SceneNodeMouseEventHandler MouseLeave;

		#endregion

		#region Private Members
		private bool _isMouseOver = false;
		#endregion

		#region Internal Members
		internal List<Point3D> RestVertexPositions { get; private set; }
		internal List<Vector3D> RestNormals { get; private set; }
		internal bool MouseHasEntered { get; set; }
		#endregion

		#region Properties
		private GeometryModel3D _model;
		/// <summary>
		/// Gets or sets the model content of this ModelSceneNode.
		/// </summary>
		public GeometryModel3D Model
		{
			get { return _model; }
			set
			{
				if (_model == value)
					return;

				if (value != null && value.Geometry != null && value.Geometry is MeshGeometry3D == false)
					throw new NotSupportedException("Only MeshGeometry3D is supported on a ModelSceneNode model.");

				_model = value;

				Data.Children[1] = (_model != null) ? _model : new GeometryModel3D();

				if (_model != null && _model.Geometry != null)
				{
					RestVertexPositions = new List<Point3D>(((MeshGeometry3D)_model.Geometry).Positions);
					RestNormals = new List<Vector3D>(((MeshGeometry3D)_model.Geometry).Normals);
				}
			}
		}
		#endregion

		#region Internal Methods
		internal void SetMouseOverState(bool isMouseOver, SceneNodeMouseEventArgs e)
		{
			if (_isMouseOver == isMouseOver)
				return;

			_isMouseOver = isMouseOver;

			// Check if mouse entered or left
			if (isMouseOver)
				RaiseMouseEnterEvent(e);
			else
				RaiseMouseLeaveEvent(e);
		}
		#endregion

		#region Event Invokers
		#region MouseEnter
		private void RaiseMouseEnterEvent(SceneNodeMouseEventArgs e)
		{
			if (MouseEnter != null)
			{
				foreach (var del in MouseEnter.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion

		#region MouseLeave
		private void RaiseMouseLeaveEvent(SceneNodeMouseEventArgs e)
		{
			if (MouseLeave != null)
			{
				foreach (var del in MouseLeave.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion
		#endregion
	}
}
