﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Windows;
using System.Windows.Media.Media3D;

namespace Ascend
{
	/// <summary>
	/// An abstract container for an object in the scene graph.
	/// </summary>
	public abstract class SceneNode : TransformableObject
	{
		#region Lifecycle
		/// <summary>
		/// SceneNode constructor.
		/// </summary>
		public SceneNode()
		{
			// Create children model group
			Data.Children.Add(new Model3DGroup());

			// Create child SceneNode collection. Each element will correspond
			// to a model in the children model group.
			Children = new SceneNodeCollection();

			// Listen to changes in children
			Children.CollectionChanged += Children_CollectionChanged;
		}
		#endregion

		#region Events
		/// <summary>
		/// Occurs when a mouse is clicked two or more times over a SceneNode.
		/// </summary>
		public event SceneNodeMouseButtonEventHandler MouseDoubleClick;

		/// <summary>
		/// Occurs when a mouse is clicked two or more times over a SceneNode.
		/// </summary>
		public event SceneNodeMouseButtonEventHandler PreviewMouseDoubleClick;

		/// <summary>
		/// Occurs when a mouse button is pressed over a SceneNode.
		/// </summary>
		public event SceneNodeMouseButtonEventHandler MouseDown;

		/// <summary>
		/// Occurs when a mouse button is pressed over a SceneNode.
		/// </summary>
		public event SceneNodeMouseButtonEventHandler PreviewMouseDown;

		/// <summary>
		/// Occurs when a mouse button is released over a SceneNode.
		/// </summary>
		public event SceneNodeMouseButtonEventHandler MouseUp;

		/// <summary>
		/// Occurs when a mouse button is released over a SceneNode.
		/// </summary>
		public event SceneNodeMouseButtonEventHandler PreviewMouseUp;

		/// <summary>
		/// Occurs when a mouse is moved while over a SceneNode.
		/// </summary>
		public event SceneNodeMouseEventHandler MouseMove;

		/// <summary>
		/// Occurs when a mouse is moved while over a SceneNode.
		/// </summary>
		public event SceneNodeMouseEventHandler PreviewMouseMove;

		/// <summary>
		/// Occurs when a mouse wheel is moved while over a SceneNode.
		/// </summary>
		public event SceneNodeMouseWheelEventHandler MouseWheel;

		/// <summary>
		/// Occurs when a mouse wheel is moved while over a SceneNode.
		/// </summary>
		public event SceneNodeMouseWheelEventHandler PreviewMouseWheel;
		#endregion

		#region Private Members
		internal Model3DCollection VisualChildren
		{
			get { return ((Model3DGroup)Data.Children[0]).Children; }
		}
		#endregion

		#region Properties
		private SceneNode _parent;
		/// <summary>
		/// Gets or sets this SceneNode's parent SceneNode.
		/// </summary>
		public SceneNode Parent
		{
			get { return _parent; }
			set
			{
				// Do nothing in this case
				if (_parent == value)
					return;

				// Remove from curent parent
				if (value == null)
				{
					_parent.Children.Remove(this);  // Will set _parent = null
					return;
				}

				// Attach to new parent (will handle removing from current parent)
				value.Children.Add(this);  // Will set _parent = value
			}
		}

		/// <summary>
		/// Gets this SceneNode's collection of child SceneNodes.
		/// </summary>
		public SceneNodeCollection Children { get; private set; }

		/// <summary>
		/// Gets the Bone to which this SceneNode is attached.
		/// This value is set when this SceneNode is added to
		/// a Bone's RigidAttachments collection.
		/// </summary>
		public Bone BoneParent { get; internal set; }

		/// <summary>
		/// Gets the TransformableObject to which this SceneNode is relatively transformed.
		/// It will be equal to Parent, BoneParent, or null.
		/// </summary>
		public override TransformableObject TransformParent
		{
			get
			{
				if (Parent != null)
					return Parent;
				else if (BoneParent != null)
					return BoneParent;
				else
					return null;
			}
		}

		/// <summary>
		/// Gets a read-only list of TransformableObjects that are transformed relative to this SceneNode.
		/// </summary>
		public override IList<TransformableObject> TransformChildren { get { return Children.ToList<TransformableObject>().AsReadOnly(); } }
		#endregion

		#region Protected Methods
		protected virtual void CloneProperties(SceneNode clone)
		{
			clone.Name = Name;
			clone.Parent = Parent;

			ClonePose(clone);
		}

		protected virtual void CloneChildren(SceneNode clone)
		{
			for (int c = 0; c < Children.Count; c++)
				clone.Children.Add(Children[c].Clone() as SceneNode);
		}
		#endregion

		#region Event Invokers
		#region MouseDoubleClick
		internal static void RaiseMouseDoubleClickEvent(SceneNode target, SceneNodeMouseButtonEventArgs e)
		{
			target.RaiseMouseDoubleClickEvent(e);
		}

		internal void RaiseMouseDoubleClickEvent(SceneNodeMouseButtonEventArgs e)
		{
			if (MouseDoubleClick != null)
			{
				foreach (var del in MouseDoubleClick.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion

		#region PreviewMouseDoubleClick
		internal static void RaisePreviewMouseDoubleClickEvent(SceneNode target, SceneNodeMouseButtonEventArgs e)
		{
			target.RaisePreviewMouseDoubleClickEvent(e);
		}

		internal void RaisePreviewMouseDoubleClickEvent(SceneNodeMouseButtonEventArgs e)
		{
			if (PreviewMouseDoubleClick != null)
			{
				foreach (var del in PreviewMouseDoubleClick.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion

		#region MouseDown
		internal static void RaiseMouseDownEvent(SceneNode target, SceneNodeMouseButtonEventArgs e)
		{
			target.RaiseMouseDownEvent(e);
		}

		internal void RaiseMouseDownEvent(SceneNodeMouseButtonEventArgs e)
		{
			if (MouseDown != null)
			{
				foreach (var del in MouseDown.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion

		#region PreviewMouseDown
		internal static void RaisePreviewMouseDownEvent(SceneNode target, SceneNodeMouseButtonEventArgs e)
		{
			target.RaisePreviewMouseDownEvent(e);
		}

		internal void RaisePreviewMouseDownEvent(SceneNodeMouseButtonEventArgs e)
		{
			if (PreviewMouseDown != null)
			{
				foreach (var del in PreviewMouseDown.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion

		#region MouseUp
		internal static void RaiseMouseUpEvent(SceneNode target, SceneNodeMouseButtonEventArgs e)
		{
			target.RaiseMouseUpEvent(e);
		}

		internal void RaiseMouseUpEvent(SceneNodeMouseButtonEventArgs e)
		{
			if (MouseUp != null)
			{
				foreach (var del in MouseUp.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion

		#region PreviewMouseUp
		internal static void RaisePreviewMouseUpEvent(SceneNode target, SceneNodeMouseButtonEventArgs e)
		{
			target.RaisePreviewMouseUpEvent(e);
		}

		internal void RaisePreviewMouseUpEvent(SceneNodeMouseButtonEventArgs e)
		{
			if (PreviewMouseUp != null)
			{
				foreach (var del in PreviewMouseUp.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion

		#region MouseMove
		internal static void RaiseMouseMoveEvent(SceneNode target, SceneNodeMouseEventArgs e)
		{
			target.RaiseMouseMoveEvent(e);
		}

		internal void RaiseMouseMoveEvent(SceneNodeMouseEventArgs e)
		{
			if (MouseMove != null)
			{
				foreach (var del in MouseMove.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion

		#region PreviewMouseMove
		internal static void RaisePreviewMouseMoveEvent(SceneNode target, SceneNodeMouseEventArgs e)
		{
			target.RaisePreviewMouseMoveEvent(e);
		}

		internal void RaisePreviewMouseMoveEvent(SceneNodeMouseEventArgs e)
		{
			if (PreviewMouseMove != null)
			{
				foreach (var del in PreviewMouseMove.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion

		#region MouseWheel
		internal static void RaiseMouseWheelEvent(SceneNode target, SceneNodeMouseWheelEventArgs e)
		{
			target.RaiseMouseWheelEvent(e);
		}

		internal void RaiseMouseWheelEvent(SceneNodeMouseWheelEventArgs e)
		{
			if (MouseWheel != null)
			{
				foreach (var del in MouseWheel.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion

		#region PreviewMouseWheel
		internal static void RaisePreviewMouseWheelEvent(SceneNode target, SceneNodeMouseWheelEventArgs e)
		{
			target.RaisePreviewMouseWheelEvent(e);
		}

		internal void RaisePreviewMouseWheelEvent(SceneNodeMouseWheelEventArgs e)
		{
			if (PreviewMouseWheel != null)
			{
				foreach (var del in PreviewMouseWheel.GetInvocationList())
				{
					if (e.Handled)
						return;

					del.Method.Invoke(del.Target, new object[] { this, e });
				}
			}
		}
		#endregion
		#endregion

		#region Event Handlers
		private void Children_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
		{
			switch (e.Action)
			{
				case NotifyCollectionChangedAction.Add:
					HandleCollectionActionAdd(sender, e);
					break;

				case NotifyCollectionChangedAction.Remove:
					HandleCollectionActionRemove(sender, e);
					break;

				default:
					throw new NotSupportedException("NotifyCollectionChangedAction.Reset is not supported.");
			}
		}

		private void HandleCollectionActionAdd(object sender, NotifyCollectionChangedEventArgs e)
		{
			foreach (SceneNode child in e.NewItems)
			{
				// Add child's data to child model group so child will render
				VisualChildren.Add(child.Data);

				// Remove child from any bone
				if (child.BoneParent != null)
					child.BoneParent.RigidAttachments.Remove(child);  // Will set child.BoneAttachment = null

				// Remove child from any current parent
				if (child.Parent != null)
					child.Parent.Children.Remove(child);  // Will set child.Parent = null

				// Set parent
				child._parent = this;
			}
		}

		private void HandleCollectionActionRemove(object sender, NotifyCollectionChangedEventArgs e)
		{
			foreach (SceneNode child in e.OldItems)
			{
				// Remove child's data from model group so it will no longer render
				VisualChildren.Remove(child.Data);

				// Clear parent
				child._parent = null;
			}
		}
		#endregion

		#region Static Methods
		/// <summary>
		/// Loads an Ascend model as a hierarchical SceneNode.
		/// </summary>
		/// <param name="filename">The path to the model.</param>
        /// <exception cref="Exception">
        /// Thrown if a failure is encountered during load.
        /// </exception>
		public static SceneNode Load(string filename)
		{
			return Loader.Load(filename);
		}

		/// <summary>
		/// Tries to load an Ascend model and returns whether or not
		/// the load was successful.
		/// </summary>
		/// <param name="filename">The path to the model.</param>
		/// <param name="sceneNode">The loaded SceneNode (will be null if load failed).</param>
		public static bool TryLoad(string filename, out SceneNode sceneNode)
		{
			try
			{
				sceneNode = Loader.Load(filename);
				return true;
			}
			catch
			{
				sceneNode = null;
				return false;
			}
		}

		/// <summary>
		/// Loads an Ascend model as a hierarchical SceneNode.
		/// </summary>
		/// <param name="stream">The stream representation of the model.</param>
        /// <exception cref="Exception">
        /// Thrown if a failure is encountered during load.
        /// </exception>
		public static SceneNode Load(Stream stream)
		{
			return Loader.Load(stream);
		}

		/// <summary>
		/// Tries to load an Ascend model and returns whether or not
		/// the load was successful.
		/// </summary>
		/// <param name="stream">The stream representation of the model.</param>
		/// <param name="sceneNode">The loaded SceneNode (will be null if load failed).</param>
		public static bool TryLoad(Stream stream, out SceneNode sceneNode)
		{
			try
			{
				sceneNode = Loader.Load(stream);
				return true;
			}
			catch
			{
				sceneNode = null;
				return false;
			}
		}

		/// <summary>
		/// Loads an Ascend model as a hierarchical SceneNode.
		/// </summary>
		/// <param name="stream">The stream representation of the model.</param>
		/// <param name="textures">A bitmap dictionary for resolving resource textures
		/// (keys must be equivalent to texture filenames defined in model file/stream,
		/// including extensions).</param>
        /// <exception cref="Exception">
        /// Thrown if a failure is encountered during load.
        /// </exception>
		public static SceneNode Load(Stream stream, Dictionary<string, Bitmap> textures)
		{
			return Loader.Load(stream, textures);
		}

		/// <summary>
		/// Tries to load an Ascend model and returns whether or not
		/// the load was successful.
		/// </summary>
		/// <param name="stream">The stream representation of the model.</param>
		/// <param name="sceneNode">The loaded SceneNode (will be null if load failed).</param>
		/// <param name="textures">A bitmap dictionary for resolving resource textures
		/// (keys must be equivalent to texture filenames defined in model file/stream,
		/// including extensions).</param>
		public static bool TryLoad(Stream stream, out SceneNode sceneNode, Dictionary<string, Bitmap> textures)
		{
			try
			{
				sceneNode = Loader.Load(stream, textures);
				return true;
			}
			catch
			{
				sceneNode = null;
				return false;
			}
		}

		/// <summary>
		/// Loads an Ascend model as a hierarchical SceneNode.
		/// </summary>
		/// <param name="stream">The stream representation of the model.</param>
		/// <param name="textureStreams">A stream dictionary for resolving embedded textures
		/// (keys must be equivalent to texture filenames defined in model file/stream,
		/// including extensions).</param>
        /// <exception cref="Exception">
        /// Thrown if a failure is encountered during load.
        /// </exception>
		public static SceneNode Load(Stream stream, Dictionary<string, Stream> textureStreams)
		{
			return Loader.Load(stream, textureStreams);
		}

		/// <summary>
		/// Tries to load an Ascend model and returns whether or not
		/// the load was successful.
		/// </summary>
		/// <param name="stream">The stream representation of the model.</param>
		/// <param name="sceneNode">The loaded SceneNode (will be null if load failed).</param>
		/// <param name="textureStreams">A stream dictionary for resolving embedded textures
		/// (keys must be equivalent to texture filenames defined in model file/stream,
		/// including extensions).</param>
		public static bool TryLoad(Stream stream, out SceneNode sceneNode, Dictionary<string, Stream> textureStreams)
		{
			try
			{
				sceneNode = Loader.Load(stream, textureStreams);
				return true;
			}
			catch
			{
				sceneNode = null;
				return false;
			}
		}
		#endregion
	}
}
