﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media.Media3D;

namespace Ascend
{
	/// <summary>
	/// Abstract implementation of an object that can be positioned, rotated, and scaled
	/// in three dimensions.
	/// </summary>
	public abstract class TransformableObject : DependencyObject
	{
		#region Lifecycle
		public TransformableObject()
		{
			Data = new Model3DGroup();

			SetupTransforms();
		}
		#endregion

		#region Private Members
		private TranslateTransform3D _translateTransform;
		private RotateTransform3D _rotateTransformX;
		private RotateTransform3D _rotateTransformY;
		private RotateTransform3D _rotateTransformZ;
		private ScaleTransform3D _scaleTransform;
		#endregion

		#region Internal Members
		/// <summary>
		/// The container to which this TransformableObject's transformations are applied.
		/// </summary>
		internal Model3DGroup Data
		{
			get { return (Model3DGroup)GetValue(DataProperty); }
			set { SetValue(DataProperty, value); }
		}

		private static readonly DependencyProperty DataProperty =
			DependencyProperty.Register("Data", typeof(Model3DGroup), typeof(TransformableObject));
		#endregion

		#region Dependency Properties
		/// <summary>
		/// Gets or sets this TransformableObject's X coordinate.
		/// </summary>
		public double X
		{
			get { return (double)GetValue(XProperty); }
			set { SetValue(XProperty, value); }
		}

		public static readonly DependencyProperty XProperty =
			DependencyProperty.Register("X", typeof(double), typeof(TransformableObject),
			new PropertyMetadata(0.0, new PropertyChangedCallback(OnXChanged)));

		private static void OnXChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
		{
			TransformableObject me = sender as TransformableObject;

			if (me == null)
				return;

			me._translateTransform.OffsetX = (double)e.NewValue;
		}


		/// <summary>
		/// Gets or sets this TransformableObject's Y coordinate.
		/// </summary>
		public double Y
		{
			get { return (double)GetValue(YProperty); }
			set { SetValue(YProperty, value); }
		}

		public static readonly DependencyProperty YProperty =
			DependencyProperty.Register("Y", typeof(double), typeof(TransformableObject),
			new PropertyMetadata(0.0, new PropertyChangedCallback(OnYChanged)));

		private static void OnYChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
		{
			TransformableObject me = sender as TransformableObject;

			if (me == null)
				return;

			me._translateTransform.OffsetY = (double)e.NewValue;
		}


		/// <summary>
		/// Gets or sets this TransformableObject's Z coordinate.
		/// </summary>
		public double Z
		{
			get { return (double)GetValue(ZProperty); }
			set { SetValue(ZProperty, value); }
		}

		public static readonly DependencyProperty ZProperty =
			DependencyProperty.Register("Z", typeof(double), typeof(TransformableObject),
			new PropertyMetadata(0.0, new PropertyChangedCallback(OnZChanged)));

		private static void OnZChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
		{
			TransformableObject me = sender as TransformableObject;

			if (me == null)
				return;

			me._translateTransform.OffsetZ = (double)e.NewValue;
		}


		/// <summary>
		/// Gets or sets this TransformableObject's X-axis angle (roll) in degrees
		/// (Rotation convention is yaw-pitch-roll / YZX.)
		/// </summary>
		public double AngleX
		{
			get { return (double)GetValue(AngleXProperty); }
			set { SetValue(AngleXProperty, value); }
		}

		public static readonly DependencyProperty AngleXProperty =
			DependencyProperty.Register("AngleX", typeof(double), typeof(TransformableObject),
			new PropertyMetadata(0.0, new PropertyChangedCallback(OnAngleXChanged)));

		private static void OnAngleXChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
		{
			TransformableObject me = sender as TransformableObject;

			if (me == null)
				return;

			((AxisAngleRotation3D)me._rotateTransformX.Rotation).Angle = (double)e.NewValue;
		}


		/// <summary>
		/// Gets or sets this TransformableObject's Y-axis angle (yaw) in degrees
		/// (Rotation convention is yaw-pitch-roll / YZX.)
		/// </summary>
		public double AngleY
		{
			get { return (double)GetValue(AngleYProperty); }
			set { SetValue(AngleYProperty, value); }
		}

		public static readonly DependencyProperty AngleYProperty =
			DependencyProperty.Register("AngleY", typeof(double), typeof(TransformableObject),
			new PropertyMetadata(0.0, new PropertyChangedCallback(OnAngleYChanged)));

		private static void OnAngleYChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
		{
			TransformableObject me = sender as TransformableObject;

			if (me == null)
				return;

			((AxisAngleRotation3D)me._rotateTransformY.Rotation).Angle = (double)e.NewValue;
		}


		/// <summary>
		/// Gets or sets this TransformableObject's Z-axis angle (pitch) in degrees
		/// (Rotation convention is yaw-pitch-roll / YZX.)
		/// </summary>
		public double AngleZ
		{
			get { return (double)GetValue(AngleZProperty); }
			set { SetValue(AngleZProperty, value); }
		}

		public static readonly DependencyProperty AngleZProperty =
			DependencyProperty.Register("AngleZ", typeof(double), typeof(TransformableObject),
			new PropertyMetadata(0.0, OnAngleZChanged));

		private static void OnAngleZChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
		{
			TransformableObject me = sender as TransformableObject;

			if (me == null)
				return;

			((AxisAngleRotation3D)me._rotateTransformZ.Rotation).Angle = (double)e.NewValue;
		}


		/// <summary>
		/// Gets or sets this TransformableObject's X-axis scale factor.
		/// </summary>
		public double ScaleX
		{
			get { return (double)GetValue(ScaleXProperty); }
			set { SetValue(ScaleXProperty, value); }
		}

		public static readonly DependencyProperty ScaleXProperty =
			DependencyProperty.Register("ScaleX", typeof(double), typeof(TransformableObject),
			new PropertyMetadata(1.0, new PropertyChangedCallback(OnScaleXChanged)));

		private static void OnScaleXChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
		{
			TransformableObject me = sender as TransformableObject;

			if (me == null)
				return;

			me._scaleTransform.ScaleX = (double)e.NewValue;
		}


		/// <summary>
		/// Gets or sets this TransformableObject's Y-axis scale factor.
		/// </summary>
		public double ScaleY
		{
			get { return (double)GetValue(ScaleYProperty); }
			set { SetValue(ScaleYProperty, value); }
		}

		public static readonly DependencyProperty ScaleYProperty =
			DependencyProperty.Register("ScaleY", typeof(double), typeof(TransformableObject),
			new PropertyMetadata(1.0, new PropertyChangedCallback(OnScaleYChanged)));

		private static void OnScaleYChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
		{
			TransformableObject me = sender as TransformableObject;

			if (me == null)
				return;

			me._scaleTransform.ScaleY = (double)e.NewValue;
		}


		/// <summary>
		/// Gets or sets this TransformableObject's Z-axis scale factor.
		/// </summary>
		public double ScaleZ
		{
			get { return (double)GetValue(ScaleZProperty); }
			set { SetValue(ScaleZProperty, value); }
		}

		public static readonly DependencyProperty ScaleZProperty =
			DependencyProperty.Register("ScaleZ", typeof(double), typeof(TransformableObject),
			new PropertyMetadata(1.0, new PropertyChangedCallback(OnScaleZChanged)));

		private static void OnScaleZChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
		{
			TransformableObject me = sender as TransformableObject;

			if (me == null)
				return;

			me._scaleTransform.ScaleZ = (double)e.NewValue;
		}
		#endregion

		#region Properties
		/// <summary>
		/// Gets or sets the name of this TransformableObject.
		/// </summary>
		public string Name { get; set; }

		/// <summary>
		/// Gets the TransformableObject to which this TransformableObject is relatively transformed.
		/// </summary>
		public abstract TransformableObject TransformParent { get; }

		/// <summary>
		/// Gets a read-only list of TransformableObjects that are transformed relative to this TransformableObject.
		/// </summary>
		public abstract IList<TransformableObject> TransformChildren { get; }

		/// <summary>
		/// Gets this TransformableObject's transform relative to its transform parent.
		/// </summary>
		public Transform3D LocalTransform { get { return Data.Transform.Clone(); } }

		/// <summary>
		/// Gets this TransformableObject's transform relative to its furthest ancestor.
		/// </summary>
		public Transform3D GlobalTransform
		{
			get
			{
				Transform3DGroup ret = new Transform3DGroup();
				ret.Children.Add(LocalTransform);

				if (TransformParent == null)
					return ret;

				ret.Children.Add(TransformParent.GlobalTransform);

				return ret;
			}
		}
		#endregion

		#region Public Methods
		/// <summary>
		/// Returns a read-only list of this TransformableObject's ancestor TransformableObjects.
		/// </summary>
		public IList<TransformableObject> GetAncestors()
		{
			List<TransformableObject> ancestors = new List<TransformableObject>();

			if (TransformParent != null)
			{
				ancestors.Add(TransformParent);
				ancestors.AddRange(TransformParent.GetAncestors());
			}

			return ancestors.AsReadOnly();
		}

		/// <summary>
		/// Returns a read-only list of this TransformableObject's descendent TransformableObjects.
		/// </summary>
		public IList<TransformableObject> GetDescendants()
		{
			List<TransformableObject> descendants = new List<TransformableObject>();

			foreach (var child in TransformChildren)
			{
				descendants.Add(child);
				descendants.AddRange(child.GetDescendants());
			}

			return descendants.AsReadOnly();
		}

		/// <summary>
		/// Creates a deep copy of this TransformableObject.
		/// </summary>
		public abstract object Clone();

		/// <summary>
		/// Returns this TransformableObject as a string.
		/// </summary>
		public override string ToString()
		{
			return (string.IsNullOrWhiteSpace(Name) ? string.Empty : Name + " ")
				+ "(" + GetType().Name + ")";
		}
		#endregion

		#region Protected Methods
		protected virtual void ClonePose(TransformableObject clone)
		{
			clone.X = X;
			clone.Y = Y;
			clone.Z = Z;

			clone.AngleX = AngleX;
			clone.AngleY = AngleY;
			clone.AngleZ = AngleZ;

			clone.ScaleX = ScaleX;
			clone.ScaleY = ScaleY;
			clone.ScaleZ = ScaleZ;
		}
		#endregion

		#region Private Methods
		private void SetupTransforms()
		{
			// Setup translation and scale transforms
			_translateTransform = new TranslateTransform3D();
			_scaleTransform = new ScaleTransform3D();

			// Setup rotation transform using YZX Euler angles
			Transform3DGroup rotateTransform = new Transform3DGroup();
			_rotateTransformX = new RotateTransform3D(new AxisAngleRotation3D(new Vector3D(1, 0, 0), 0));
			_rotateTransformY = new RotateTransform3D(new AxisAngleRotation3D(new Vector3D(0, 1, 0), 0));
			_rotateTransformZ = new RotateTransform3D(new AxisAngleRotation3D(new Vector3D(0, 0, 1), 0));
			rotateTransform.Children.Add(_rotateTransformY); // => yaw
			rotateTransform.Children.Add(_rotateTransformZ); // => pitch
			rotateTransform.Children.Add(_rotateTransformX); // => roll

			// Setup total transform (order of children matters)
			Data.Transform = new Transform3DGroup();
			((Transform3DGroup)Data.Transform).Children.Add(_scaleTransform);
			((Transform3DGroup)Data.Transform).Children.Add(rotateTransform);
			((Transform3DGroup)Data.Transform).Children.Add(_translateTransform);
		}
		#endregion
	}
}
