﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Input;
using Ascend;

namespace AscendViewer
{
	public class AnimatedContentNodeViewModel : ModelContentNodeViewModel
	{
		#region Lifecycle
		public AnimatedContentNodeViewModel(AnimatedSceneNode animatedSceneNode, ContentExplorerViewModel container)
			: base(animatedSceneNode, container)
		{
			Animations = animatedSceneNode.ListAnimations();

			if (HasAnimations)
				AnimationIndex = 0;

			if (animatedSceneNode.Armature != null)
				Children.Insert(0, Create(animatedSceneNode.Armature, container));
			
			PreviousAnimationCommand = new InterfaceCommand(() =>
				{
					AnimationIndex--;
				},
				() => Animations.Count > 1 && IsPaused
			);

			NextAnimationCommand = new InterfaceCommand(() =>
				{
					AnimationIndex++;
				},
				() => Animations.Count > 1 && IsPaused
			);
		}
		#endregion

		#region Properties
		public ICommand PreviousAnimationCommand { get; private set; }
		public ICommand NextAnimationCommand { get; private set; }

		public List<string> Animations { get; private set; }

		public bool HasAnimations { get { return Animations.Count > 0; } }

		private int _animationIndex = -1;
		public int AnimationIndex
		{
			get { return _animationIndex; }
			set
			{
				if (_animationIndex == value)
					return;

				_animationIndex = value;

				// Wrap-around
				if (_animationIndex < 0)
					_animationIndex = Animations.Count + _animationIndex;
				else if (_animationIndex >= Animations.Count)
					_animationIndex = Animations.Count - _animationIndex;

				CurrentAnimation = Animations[_animationIndex];
			}
		}

		private string _currentAnimation;
		public string CurrentAnimation
		{
			get { return _currentAnimation; }
			set
			{
				if (_currentAnimation == value)
					return;

				_currentAnimation = value;
				NotifyPropertyChanged("CurrentAnimation");
			}
		}

		private double _fps = 60.0;
		public double FPS
		{
			get { return _fps; }
			set
			{
                if (value == 0.0)
                    throw new ArgumentException("Invalid frame rate");

				if (_fps == value)
					return;

				_fps = value;
				NotifyPropertyChanged("FPS");
			}
		}

		private bool _isPlaying;
		public bool IsPlaying
		{
			get { return _isPlaying; }
			set
			{
				if (_isPlaying == value || Data == null)
					return;

				_isPlaying = value;

				if (_isPlaying)
					((AnimatedSceneNode)Data).Play(CurrentAnimation, true, false, FPS);
				else
					((AnimatedSceneNode)Data).Pause();

				NotifyPropertyChanged("IsPlaying");
				NotifyPropertyChanged("IsPaused");
			}
		}

		public bool IsPaused { get { return !IsPlaying; } }
		#endregion
	}
}
