﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Input;
using Ascend;

namespace AscendViewer
{
	public abstract class ContentNodeViewModel : ObservableObject
	{
		#region Lifecycle
		public ContentNodeViewModel(TransformableObject data, ContentExplorerViewModel container)
		{
			Data = data;

			Children = new List<ContentNodeViewModel>();

			SelectCommand = new InterfaceCommand(() => container.SelectedContentNode = this);

            LookAtCommand = new InterfaceCommand(() =>
            {
				container.Scene.LookAt(Data);
            });
		}
		#endregion

        #region Properties
        public ICommand SelectCommand { get; private set; }
        public ICommand LookAtCommand { get; protected set; }

		public TransformableObject Data { get; private set; }

		public string Name { get; protected set; }

		public List<ContentNodeViewModel> Children { get; protected set; }

		public virtual string SelfAsString { get { return Data.ToString(); } }

		public Type Type { get { return GetType(); } }

		private bool _isSelected = false;
		public bool IsSelected
		{
			get { return _isSelected; }
			set
			{
				if (_isSelected == value)
					return;

				_isSelected = value;
				NotifyPropertyChanged("IsSelected");
			}
		}

		private bool _isExpanded = true;
		public bool IsExpanded
		{
			get { return _isExpanded; }
			set
			{
				if (_isExpanded == value)
					return;

				_isExpanded = value;
				NotifyPropertyChanged("IsExpanded");
			}
		}

		public bool CanExpand { get { return Children.Count > 0; } }
		#endregion

		#region Public Methods
		public List<ContentNodeViewModel> SelfAndDescendants()
		{
			List<ContentNodeViewModel> list = new List<ContentNodeViewModel>();

			list.Add(this);

			foreach (var child in Children)
				list.AddRange(child.SelfAndDescendants());

			return list;
		}
		#endregion

		#region Protected Methods
		protected void InitializeChildren(IEnumerable<TransformableObject> children, ContentExplorerViewModel container)
		{
			foreach (var child in children)
				Children.Add(Create(child, container));
		}
		#endregion

		#region Static Methods
		public static ContentNodeViewModel Create(TransformableObject obj, ContentExplorerViewModel container)
		{
			if (obj is AnimatedSceneNode)
				return new AnimatedContentNodeViewModel((AnimatedSceneNode)obj, container);
			else if (obj is ModelSceneNode)
				return new ModelContentNodeViewModel((ModelSceneNode)obj, container);
			else if (obj is Bone)
				return new BoneContentNodeViewModel((Bone)obj, container);
			else
				throw new NotSupportedException("Unknown content node encountered.");
		}

		public static ContentNodeViewModel Create(Armature armature, ContentExplorerViewModel container)
		{
			return new ArmatureContentNodeViewModel(armature, container);
		}
		#endregion
	}
}
