﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Forms;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Media3D;
using Ascend;

namespace AscendViewer
{
	public class MainWindowViewModel : ObservableObject
	{
		#region Lifecycle
		public MainWindowViewModel(Window window)
		{
			_window = window;

			OpenFileCommand = new InterfaceCommand(OpenFileAction);
			OpenHelpCommand = new InterfaceCommand(OpenHelpAction);

			Scene = new Scene3D();
			Scene.Camera.Position = new Point3D(0, 3, 5);
			Scene.Camera.LookDirection = new Vector3D(0, -3, -5);
			Scene.RenderRoot.Children.Add(new LightSceneNode(
				new DirectionalLight(Colors.White, new Vector3D(-1, -2, -3))));
			Scene.RenderRoot.Children.Add(new LightSceneNode(
				new DirectionalLight(Colors.White, new Vector3D(1, 2, 3))));

			ContentExplorer = new ContentExplorerViewModel() { Scene = Scene };
		}
		#endregion

		#region Private Members
		private Window _window;
		#endregion

		#region Properties
		public ICommand OpenFileCommand { get; private set; }
		public ICommand OpenHelpCommand { get; private set; }

		public Scene3D Scene { get; private set; }

		public ContentExplorerViewModel ContentExplorer { get; private set; }

		private string _filePath;
		public string FilePath
		{
			get { return _filePath; }
			set
			{
				if (_filePath == value)
					return;

				_filePath = value;
				NotifyPropertyChanged("FilePath");
			}
		}
		#endregion

		#region Public Methods
		public void OpenFile(string filePath)
		{
			if (Scene.RenderRoot.Children.Count() > 2)
				Scene.RenderRoot.Children.RemoveAt(2);

			ContentExplorer.ContentNode = null;
			ContentExplorer.FileName = string.Empty;
			FilePath = string.Empty;

			SceneNode node;
			if (SceneNode.TryLoad(filePath, out node))
			{
				FilePath = filePath;

				Scene.RenderRoot.Children.Add(node);
				ContentExplorer.ContentNode = ContentNodeViewModel.Create(node, ContentExplorer);
				ContentExplorer.SelectedContentNode = ContentExplorer.ContentNode;
				ContentExplorer.IsExpanded = true;
				ContentExplorer.FileName = Path.GetFileName(FilePath);

				// Subscribe to clicks to select node
				foreach (var child in Scene.RenderRoot.GetDescendants().OfType<SceneNode>())
					child.MouseDoubleClick += SceneNode_MouseDoubleClick;
			}
			else
			{
				System.Windows.MessageBox.Show("Unable to open file \'" + Path.GetFileName(filePath) + "\'.",
					"Open", MessageBoxButton.OK, MessageBoxImage.Warning);
			}
		}
		#endregion

		#region Private Methods
		private void OpenFileAction()
		{
			OpenFileDialog dlg = new OpenFileDialog();
			dlg.Filter = "Ascend (*.asc)|*.asc";
			dlg.RestoreDirectory = true;

			if (dlg.ShowDialog() == DialogResult.OK)
				OpenFile(dlg.FileName);
		}

		private void OpenHelpAction()
		{
			HelpWindow wnd = new HelpWindow();
			wnd.Owner = _window;

			wnd.Show();
		}
		#endregion

		#region Event Handlers
		void SceneNode_MouseDoubleClick(object sender, SceneNodeMouseButtonEventArgs e)
		{
			// Get clicked SceneNode
			SceneNode node = sender as SceneNode;

			// Find associated ContentNodeViewModel and select it
			ContentExplorer.SelectedContentNode = ContentExplorer.ContentNode.SelfAndDescendants().Find(c => c.Data == node);

			// Point camera to node
			Scene.LookAt(node);

			// Handle event so ancestors are not selected
			e.Handled = true;
		}

		#endregion
	}
}
