﻿/*
	Copyright (c) 2013 Andrew Meyer

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
	associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute,
	sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial
	portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Windows.Input;

namespace Demo04_BasicAnimation
{
	/// <summary>
	/// A class for handling events on UI elements
	/// </summary>
	public class InterfaceCommand : ICommand
	{
		#region Properties
		readonly Action _execute;
		readonly Func<bool> _canExecute;
		#endregion

		#region ICommand Members
		/// <summary>
		/// Event handler for when validation function changes
		/// </summary>
		public event EventHandler CanExecuteChanged
		{
			add
			{
				if (_canExecute != null)
					CommandManager.RequerySuggested += value;
			}

			remove
			{
				if (_canExecute != null)
					CommandManager.RequerySuggested -= value;
			}
		}
		#endregion

		#region Lifecycle
		/// <summary>
		/// Creates a command that will perform a given action when it is executed
		/// </summary>
		/// <param name="execute">Action to perform</param>
		/// <param name="canExecute">Function to validate execution ability</param>
		public InterfaceCommand(Action execute, Func<bool> canExecute = null)
		{
			if (execute == null)
				throw new ArgumentNullException("execute");

			_execute = execute;
			_canExecute = canExecute;
		}
		#endregion

		#region ICommand Methods
		/// <summary>
		/// Performs this command's action
		/// </summary>
		/// <param name="parameter">(Unused)</param>
		public virtual void Execute(object parameter = null)
		{
			_execute();
		}

		/// <summary>
		/// Validates execution ability
		/// </summary>
		/// <param name="parameter">(Unused)</param>
		/// <returns>Whether or not this command's action can be performed</returns>
		public virtual bool CanExecute(object parameter = null)
		{
			return _canExecute == null ? true : _canExecute();
		}
		#endregion
	}
}